import os, glob
import numpy

# _default_NDDEM_path = '~/Dropbox/Research/Codes/DEM_ND/Samples/radioSphere'
_default_NDDEM_path = '~/code/NDDEM/Samples/radioSphere'

def write_infile(xyzMM,radiiMM,nt,dumpfolder=_default_NDDEM_path):

    files = glob.glob(f'{os.path.expanduser(dumpfolder)}/*')
    for f in files: os.remove(f)
    with open(f"{os.path.expanduser(dumpfolder)}/in",'w') as f:
        f.write(f"dimensions 3 {len(xyzMM)}\nradius -1 {radiiMM[0]}\nboundary 0 PBC  {xyzMM[:,0].min()} {xyzMM[:,0].max()}\nboundary 1 PBC  {xyzMM[:,1].min()} {xyzMM[:,1].max()}\nboundary 2 PBC  {xyzMM[:,2].min()} {xyzMM[:,2].max()}\nset T {nt}\nset dt 1\nset tdump 1\nEOF")

def write_dumpfile(xyzMM,radiiMM,iteration,dumpfolder=_default_NDDEM_path):
    numpy.savetxt(f"{os.path.expanduser(dumpfolder)}/dump-{iteration:05d}.csv",
                  numpy.hstack([xyzMM,numpy.expand_dims(radiiMM,1)]),
                  delimiter=',',
                  header="x0,x1,x2,R",
                  comments="")



import numpy as np
from scipy.spatial.distance import cdist

def DEM_step(locMMin, radiiMM, k=0.5):
    """
    Lightweight DEM with assumption of same size spheres as mechanical regularisation

    Parameters
    ----------
        locMMin : Nx3 2D numpy array of floats
            xyz positions of spheres in mm, with the origin being the middle of the detector

        radiiMM : 1D numpy array of floats
            Particle radii for projection

        k : float, optional
            Stiffness and timestep wrapped into one
            Default = 0.1

    Returns
    -------
        locMMin : output positions
    """
    locMM = locMMin.copy()

    if radiiMM.min() != radiiMM.max():
        print("DEM.DEM_step(): WARINING I assume all radii are the same, taking first one")
    #k = 0.1 # stiffness and timestep wrapped into one
    np = len(locMM)
    delta = cdist(locMM,locMM) - 2*radiiMM[0] # assuming all radii the same
    # print(delta)
    for i in range(0,np):
        for j in range(i+1,np):
            if delta[i,j] < 0:
                # print(i,j,i+j)
                branch_vector = locMM[i] - locMM[j]
                F = -k*delta[i,j]*branch_vector
                locMM[i] += F
                locMM[j] -= F
    return locMM


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    # locMM = np.array([[0,0,0],
    #                   [0,1.8,0],
    #                   [0,0,3],
    #                   [3,0,0]])
    locMM = np.random.rand(50,3)*20
    print(len(locMM))
    radiiMM = 1.*np.ones(len(locMM))
    for t in range(100):
        locMM = DEM_step(locMM,radiiMM)
        plt.ion()
        plt.plot(locMM[:,0],locMM[:,1],'.')
        # plt.show()
        plt.pause(0.01)
    #print(locMM)



import numpy, scipy, tifffile
import matplotlib.pyplot as plt
import scipy.optimize

#import PySide2
#import pyqtgraph as pg

#from functions import *
import time

numpy.set_printoptions(suppress=True)

def optimiseSensitivityFields(radioMM, xyzGuessesMM, radiiMM, iterationsMax=20, minDeltaMM=0.01, perturbationMM=None, sourceDetectorDistMM=100, pixelSizeMM=0.1, detectorResolution=[512,512], verbose=0, DEMcorr=False, GRAPH=False, NDDEM_output=False, blur=None, showConvergence=False, optimiseGL=False, optimiseGLcalib=None):
    """
    This function takes in a reference projection (radioMM) in mm and sphere positions
    and moves them around to minimise the residual as far as possible.
    The technique used is the calculation of a sensitivity field in x, y, z with a perturbation of the
    synthetic radiograph.
    This is done locally in a ROI around each particle, which avoids doing the projection for all pixels.

    Parameters
    ----------
        radioMM : 2D numpy array of floats
            Projection of distance crossed in spheres in MM

        xyzGuessesMM : 2D numpy array of floats
            Nx3 guesses for the positions of the particles in MM

        radiiMM : 1D numpy array of floats
            Particle radii for projection

        iterationsMax : int, optional
            Number of iterations to perform
            Default = 20

        minDeltaMM : float, optional
            If step size gets below this value, stop
            Default = 0.01

        perturbationMM : either a single float or a 3-component list of floats, optional
            Size of the perturbation step to compute the sensitivity field
            Default = mean particle radius * numpy.array([SOD^2/(2*SDD), 0.5, 0.5])

        sourceDetectorDistMM : float, optional
            Distance between x-ray source and middle of detector
            Default = 100

        pixelSizeMM : float, optional
            Pixel size on detector in mm
            Default = 0.1

        detectorResolution : 2-component list of ints, optional
            Number of pixels height, width of detector
            Default = [512,512]

        #projector : string, optional
            #Algorithm for the projector
            #Default = 'C'

        verbose : int, optional
            Print noisly? 0 = quiet except warning, 1=iterations, 2=everything
            Default = 1

        DEMcorr : bool, optional
            Perform a soft DEM correction between all iterations?
            Default = False

        NDDEM_output : bool, optional
            Save every iteration so that it can be viewed using NDDEM - sorry Eddy

        blur : float, optional
            sigma of blur to pass to scipy.ndimage.gaussian_filter to
            blur the radiograph at the end of everything

        showConvergence : bool, optional
            show a graph of the system converging with iterations
            Default = False

        optimiseGL : bool, optional
            Perform optimisation on image in greylevels rather than mm?
            Default = False

        optimiseGLcalib : dictionary, optional
            Calibration to apply for converting mm to greylevels and back
            Default = None


    Returns
    -------
        xzyMM : Corrected positions array
    """
    import radioSphere.projectSphere

    # This is the function that will be optimised to match synthetic residuals to the real measured ones
    # ...it's the heart of the approach
    # attempt to find weights a, b, c for the x, y, z sensitivity fields with leastsquares
    def optimiseMe(weights, residualToMatch, sensXYZ):
        #print("optimiseMe(): weights = ", weights)
        output = residualToMatch - weights[0]*sensXYZ[0] - weights[1]*sensXYZ[1] - weights[2]*sensXYZ[2]
        #return numpy.ravel(output[output != 0])
        return numpy.ravel(output)

    if optimiseGL: radio = radioSphere.mm2gl(radioMM, calib=optimiseGLcalib)
    else:          radio = radioMM

    if GRAPH:
        #from PyQt5 import QtGui
        #app = QtGui.QApplication([])
        #w = QtGui.QWidget()
        plt.subplot(1,1,1)
        plt.axis([radioMM.shape[1], 0, radioMM.shape[0], 0])
        plt.ion()
        ##plt.show()
        ##plt.subplot(1,1,1)
        #imv = pg.ImageView()
        ##imv.show()
        #layout = QtGui.QGridLayout()
        #w.setLayout(layout)
        #layout.addWidget(imv, 0, 0)
        #w.show()
        #app.exec_()
        #pg.GraphicsWindow()
        #pw = pg.image(radioMM)
        pass

    if NDDEM_output:
        from radioSphere.DEM.nddem import write_infile
        write_infile(xyzGuessesMM,radiiMM,iterationsMax)

    assert(len(radiiMM) == xyzGuessesMM.shape[0]), "optimiseSensitivityFields(): number of radii and sphere positions not the same"

    sourceObjectDistanceMM = numpy.mean(numpy.sqrt(numpy.sum(xyzGuessesMM**2,axis=1)))

    # perturbXscaling = 4
    # perturbXscaling = sourceObjectDistanceMM**2/radiiMM[0]/sourceDetectorDistMM
    perturbXscaling = sourceObjectDistanceMM/radiiMM[0]

    if perturbationMM is None:
        # Pixel size in middle of sample
        pixelSizeMMatCOR = pixelSizeMM * (numpy.mean(xyzGuessesMM[:,0]) / sourceDetectorDistMM )

        # Perturb 1px in detector and 4 in the x-direction (this should be function of beam angle)
        perturbationMM = (perturbXscaling*pixelSizeMMatCOR, pixelSizeMMatCOR, pixelSizeMMatCOR)
        print(f"optimiseSensitivityFields(): using a perturbation of (which is 1 px in detector plane and larger in X):\n\tX: {perturbationMM[0]:0.3f}mm Y: {perturbationMM[1]:0.3f}mm Z: {perturbationMM[2]:0.3f}mm")
    elif isinstance(perturbationMM, float):
        perturbationMM = (perturbXscaling*perturbationMM, perturbationMM, perturbationMM)
        print(f"optimiseSensitivityFields(): using a perturbation of:\n\tX: {perturbationMM[0]:0.3f}mm Y: {perturbationMM[1]:0.3f}mm Z: {perturbationMM[2]:0.3f}mm")
    # initialise variables
    iterations = 0
    xyzMM = xyzGuessesMM.copy().astype('<f4')
    xyzMMprev = xyzGuessesMM.copy().astype('<f4')
    if verbose > 1: print("Initial pos:\n", xyzMM)

    #step = numpy.array([numpy.inf, numpy.inf, numpy.inf])
    dX = numpy.inf
    # outputForFigure = []
    while iterations < iterationsMax and dX > minDeltaMM:
        if verbose > 0: print("\tIteration Number", iterations, end='')
        if verbose > 1: print("\tperturbationMM:\t", perturbationMM)

        # Generate radio with current guess of position (xyzMM)
        guessedRadioMM = radioSphere.projectSphere.projectSphereMM( xyzMM,
                                                                    radiiMM,
                                                                    sourceDetectorDistMM=sourceDetectorDistMM,
                                                                    pixelSizeMM=pixelSizeMM,
                                                                    detectorResolution=detectorResolution,
                                                                    blur=blur)
        if optimiseGL: guessedRadio = radioSphere.mm2gl(guessedRadioMM, calib=optimiseGLcalib)
        else:          guessedRadio = guessedRadioMM

        #print(numpy.square(residualMM).sum(), end="")

        ## create an empty mask that stores locations with particles in the ROI
        #if removeBackground: isBackgroundMask = numpy.ones_like(radioMM,dtype=bool)
        ## Could easily parallelise here, these are all independent corrections
        limits = [ [] for _ in range(len(radiiMM)) ]
        for sphere in range(len(radiiMM)):
            # if verbose > 2: print("\tSphere {} of {}".format(sphere+1, len(radiiMM)))
            # Compute sensitivity fields for this perturbationMM -- this is the current virtual radiograph
            #   Perturbed by a given perturbationMM in x, y, z independently

            # Compute sensitivity field only for this particle, this is debatable, so generate its reference projection
            # Since we're going to try to do this locally, let's pre-compute the limits on the detector
            limits[sphere] = radioSphere.projectSphere.singleSphereToDetectorPixelRange(xyzMM[sphere],
                                                                                radiiMM[sphere]*1.2,
                                                                                radiusMargin=0.1,
                                                                                sourceDetectorDistMM=sourceDetectorDistMM,
                                                                                pixelSizeMM=pixelSizeMM,
                                                                                detectorResolution=detectorResolution)

            #if removeBackground: isBackgroundMask[limits[sphere][0,0]:limits[sphere][0,1], limits[sphere][1,0]:limits[sphere][1,1]] = False

        # update residual
        residual = radio - guessedRadio

        if verbose > 1: tifffile.imsave("./residuals/residual-iteration-{:04d}.tif".format(iterations), residual.astype('f4'))

        for sphere in range(len(radiiMM)):
            if verbose > 2: print("\tSphere {} of {}".format(sphere+1, len(radiiMM)))
            # Compute local reference projection for this sphere with ROI activated around the sphere
            sphereRefProjectionMM = radioSphere.projectSphere.projectSphereMM(  numpy.array([xyzMM[sphere]]),
                                                                                numpy.array([radiiMM[sphere]]),
                                                                                sourceDetectorDistMM=sourceDetectorDistMM,
                                                                                pixelSizeMM=pixelSizeMM,
                                                                                detectorResolution=detectorResolution,
                                                                                ROIcentreMM=xyzMM[sphere].copy(),
                                                                                ROIradiusMM=radiiMM[sphere]*1.2,
                                                                                blur=blur)
            if optimiseGL: sphereRefProjection = radioSphere.mm2gl(sphereRefProjectionMM, calib=optimiseGLcalib)
            else:          sphereRefProjection = sphereRefProjectionMM

            # Pre-allocate sens field, since it's local it should be the same shape as sphereRefProjectionMM
            sensXYZ = numpy.zeros((3, sphereRefProjection.shape[0], sphereRefProjection.shape[1]), dtype=float)

            #if GRAPH and iterations%10==0:
            if GRAPH:
                #plt.clf()
                plt.subplot(1,4,1)
                # plt.ion()
                if optimiseGL:
                    plt.title(f"Current Residual (GL) iteration={iterations}")
                    plt.imshow(residual, cmap='coolwarm', vmin=radioSphere.mm2gl(-0.1), vmax=radioSphere.mm2gl(0.1))
                else:
                    plt.title(f"Current Residual (mm) LUT: [-0.1, 0.1] iteration={iterations}")
                    plt.imshow(residual, cmap='coolwarm', vmin=-0.1, vmax=0.1)
                plt.pause(0.01)



            # For each direction in X, Y, Z
            for i in range(3):
                # Perturb just one direction
                tmp = xyzMM[sphere].copy()
                tmp[i] += perturbationMM[i]
                # Here the ROI parameters are the same as the undisturbed sphere to guarantee the same size
                #   projection being returned
                perturbedProjectionMM = radioSphere.projectSphere.projectSphereMM(  numpy.array([tmp]),
                                                                                    numpy.array([radiiMM[sphere]]),
                                                                                    sourceDetectorDistMM=sourceDetectorDistMM,
                                                                                    pixelSizeMM=pixelSizeMM,
                                                                                    detectorResolution=detectorResolution,
                                                                                    ROIcentreMM=xyzMM[sphere].copy(),
                                                                                    ROIradiusMM=radiiMM[sphere]*1.2,
                                                                                    blur=blur)

                if optimiseGL: perturbedProjection = radioSphere.mm2gl(perturbedProjectionMM, calib=optimiseGLcalib)
                else:          perturbedProjection = perturbedProjectionMM

                sensXYZ[i] = sphereRefProjection - perturbedProjection
                if GRAPH:
                    plt.subplot(1,4,2+i)
                    if optimiseGL:
                        plt.title(f"Sens{'XYZ'[i]} (GL) LUT: [{radioSphere.mm2gl(-2)}, {radioSphere.mm2gl(2)}]\nSphere = {sphere}, Perturbation {perturbationMM[i]} mm")
                        plt.imshow(sensXYZ[i], cmap='coolwarm', vmin=radioSphere.mm2gl(-2), vmax=radioSphere.mm2gl(2))
                    else:
                        plt.title(f"Sens{'XYZ'[i]} (mm) LUT: [-2, 2]\nSphere = {sphere}, Perturbation {perturbationMM[i]} mm")
                        plt.imshow(sensXYZ[i], cmap='coolwarm', vmin=-2, vmax=2)
            # for making sensivitiy single particle figure in paper uncomment the next line
            # numpy.save('cache/sensXYZ.npy', sensXYZ)

            if GRAPH: plt.pause(1e-6); #plt.show()

            # Reset step which will be filled with the step in XZY combining weighted combination of sensitivity fields
            #   N.B. not to be confused with perturbationMM, which for now is not changed during iterations
            #step = numpy.zeros(3)

            LSQret = scipy.optimize.least_squares(optimiseMe,
                                                  [1.0, 1.0, 1.0],
                                                  args=[residual[limits[sphere][0,0]:limits[sphere][0,1], limits[sphere][1,0]:limits[sphere][1,1]], sensXYZ],
                                                  verbose=False,
                                                  method='lm',
                                                  diff_step=1.0)
            if LSQret['success'] == True:
                #print('LSQ Success!')
                #step = LSQret['x'] * perturbationMM[i]
                #for i in range(3):
                #    xyzMM[sphere][i] -= step[i] * perturbationMM[i]
                xyzMM[sphere] -= LSQret['x'] * perturbationMM
            else:
                print("LSQ failed to converge")

            if verbose > 1:
                #print("\t\tstep:\t",step)
                print("\t\tpos:\t",xyzMM[sphere])

        ### End optimisation iterations

        ### Now a soothing  DEM step can be applied to the current updated XYZ positions
        if DEMcorr:
            import radioSphere.DEM
            xyzMMnew = radioSphere.DEM.DEM_step(xyzMM,radiiMM)
            if verbose > 0: print("   DEM changed positions by: ", numpy.linalg.norm(xyzMM-xyzMMnew), end='')
            xyzMM = xyzMMnew

        dX = numpy.linalg.norm(xyzMM-xyzMMprev)
        # outputForFigure.append([numpy.linalg.norm(step),numpy.sqrt(numpy.sum(residualMM.flatten()**2))])
        # if verbose > 0: print("   |deltaMM|: ", numpy.linalg.norm(xyzMM-xyzMMprev), end='')
        print(f"   |deltaMM|: {dX:0.5f}", end='\r')

        xyzMMprev = xyzMM.copy()

        if NDDEM_output:
            from radioSphere.DEM.nddem import write_dumpfile
            write_dumpfile(xyzMM,radiiMM,iterations)

        if showConvergence:
            # plt.figure(2)
            plt.ion()
            plt.semilogy(iterations,numpy.sqrt(numpy.sum(residual.flatten()**2)),'k.')
            plt.xlabel('Iterations')
            if optimiseGL:
                plt.ylabel('Sum of squared residuals (GL^2)')
            else:
                plt.ylabel('Sum of squared residuals (mm^2)')
            plt.pause(1e-2)
            plt.show()

        iterations += 1

    # numpy.savetxt('./cache/optimiserSensitivityField.csv',outputForFigure,delimiter=',',header='DisplacementNorm,SquaredResidual')

    if verbose > 0:
        if dX <= minDeltaMM:
            # Check that we exited based on displacement
            print("\n\toptimiseSensitivityFields(): Got below {} in {} iterations".format(minDeltaMM, iterations))
        else:
            # Check that we exited based on displacement
            print("\n\toptimiseSensitivityFields(): hit max iterations ({})".format(iterations))

    plt.ioff()
    return xyzMM



def _optimiseSensitivityFieldsMultiProj(radioMM, xyzGuessesMM, radiiMM, transformationCentresMM, transformationMatrices, iterationsMax=20, minDeltaMM=0.01, perturbationMM=numpy.array([0.5, 0.5, 0.5]), sourceDetectorDistMM=100, pixelSizeMM=0.1, detectorResolution=[512,512], solver='leastsquares', verbose=0, DEMcorr=False, GRAPH=False):
    """
    This function takes in a series of reference projection (radioMM) in mm and sphere positions
    and moves them around to minimise the residual as far as possible.
    The technique used is the calculation of a sensitivity field in x, y, z with a perturbation of the
    synthetic radiograph.
    This is done locally in a ROI around each particle, which avoids doing the projection for all pixels.

    Parameters
    ----------
        radioMM : 3D numpy array of floats
            N Projections of distance crossed in spheres in MM, where N is the number of different views

        xyzGuessesMM : 2D numpy array of floats
            Nx3 guesses for the positions of the particles in MM

        radiiMM : 1D numpy array of floats
            Particle radii for projection

        transformationCentresMM : Nx3 numpy array
            A tranformation centre (XYZ) for each projection

        transformationMatrices : Nx3x3 numpy array
            A 3x3 transformation matrix (XYZ) for each projection

        iterationsMax : int, optional
            Number of iterations to perform
            Default = 20

        minDeltaMM : float, optional
            If step size gets below this value, stop
            Default = 0.01

        perturbationMM : 3-component list of floats, optional
            Size of the perturbation step to compute the sensitivity field
            Default = (2.0, 0.5, 0.5)

        sourceDetectorDistMM : float, optional
            Distance between x-ray source and middle of detector
            Default = 100

        pixelSizeMM : float, optional
            Pixel size on detector in mm
            Default = 0.1

        detectorResolution : 2-component list of ints, optional
            Number of pixels height, width of detector
            Default = [512,512]

        #projector : string, optional
            #Algorithm for the projector
            #Default = 'C'

        solver : string, optional
            Way in which the sensitivity fields
            are combined to minimise residual.
            Options = 'homemade', 'leastsquares'
            Default = 'leastsquares'

        verbose : int, optional
            Print noisly? 0 = quiet except warning, 1=iterations, 2=everything
            Default = 1

        DEMcorr : bool, optional
            Perform a soft DEM correction between all iterations?
            Default = False

    Returns
    -------
        xzyMM : Corrected positions array
    """
    import radioSphere.projectSphere

    assert(len(radiiMM) == xyzGuessesMM.shape[0]), "optimiseSensitivityFieldsMultiProj(): number of radii and sphere positions not the same"
    assert(radioMM.shape[0] == transformationCentresMM.shape[0])
    assert(radioMM.shape[0] == transformationMatrices.shape[0])
    if numpy.all(numpy.array(transformationCentresMM[0]) != numpy.array([0.,0.,0.])):  print("optimiseSensitivityFieldsMultiProj(): master transformation centre is not zero, do you know what you're doing?")
    if numpy.all(numpy.array(transformationMatrices[0]) != numpy.eye(3)):              print("optimiseSensitivityFieldsMultiProj(): master transformation matrix is not identity, do you know what you're doing?")

    nProj = radioMM.shape[0]

    if verbose>1: print("optimiseSensitivityFieldsMultiProj(): Number of projections = ", nProj)

    # initialise variables
    iterations = 0
    xyzMM = xyzGuessesMM.copy().astype('<f4')
    xyzMMprev = xyzGuessesMM.copy().astype('<f4')
    if verbose > 1: print("Initial pos:\n", xyzMM)

    step = numpy.array([numpy.inf, numpy.inf, numpy.inf])

    while iterations < iterationsMax and numpy.linalg.norm(step*perturbationMM) > minDeltaMM:
        if verbose > 0: print("\tIteration Number", iterations, end='')
        if verbose > 1: print("\tperturbationMM:\t", perturbationMM)
        if verbose > 1: print("\txyzMM:\t", xyzMM)

        # Generate radio stack with current guess of position (xyzMM)
        guessedRadioMM = numpy.zeros_like(radioMM)
        for n in range(nProj):
            guessedRadioMM[n] = radioSphere.projectSphere.projectSphereMM(  xyzMM,
                                                                            radiiMM,
                                                                            sourceDetectorDistMM=sourceDetectorDistMM,
                                                                            pixelSizeMM=pixelSizeMM,
                                                                            detectorResolution=detectorResolution,
                                                                            transformationCentreMM=transformationCentresMM[n],
                                                                            transformationMatrix=transformationMatrices[n])

        # update residual
        residualMM = radioMM - guessedRadioMM

        if GRAPH:
            for n in range(nProj):
                plt.subplot(nProj,5,n*5+1)
                plt.title("Proj {} Ref Projection (mm) LUT: [-2, 2]".format(n))
                plt.imshow(radioMM[n])

                plt.subplot(nProj,5,n*5+2)
                plt.title("Proj {} Current Residual (mm) LUT: [-2, 2]".format(n))
                plt.imshow(residualMM[n], cmap='coolwarm', vmin=-2, vmax=2)


        # Could easily parallelise here, these are all independent corrections
        for sphere in range(len(radiiMM)):

            if verbose > 1: print("\tSphere {} of {}".format(sphere+1, len(radiiMM)))
            # Compute sensitivity fields for this perturbationMM -- this is the current virtual radiograph
            #   Perturbed by a given perturbationMM in x, y, z independently

            # This better be a list, since (due to different zooms) the ROIs might be different sizes for the different projections
            sensXYZall = []
            # This is a list of ROI residuals built here in order not to need to export limits and to avoid
            #   passing whole residual to the optimisation function below
            residualMMroi  = []
            for n in range(nProj):
                # Compute sensitivity field only for this particle, this is debatable, so generate its reference projection
                # Since we're going to try to do this locally, let's pre-compute the limits on the detector
                limits = radioSphere.projectSphere.singleSphereToDetectorPixelRange(xyzMM[sphere],
                                                                                    radiiMM[sphere]*1.2,
                                                                                    radiusMargin=0.1,
                                                                                    sourceDetectorDistMM=sourceDetectorDistMM,
                                                                                    pixelSizeMM=pixelSizeMM,
                                                                                    detectorResolution=detectorResolution,
                                                                                    transformationCentreMM=transformationCentresMM[n],
                                                                                    transformationMatrix=transformationMatrices[n])

                # Compute local reference projection for this sphere with ROI activated around the sphere
                sphereRefProjectionMM = radioSphere.projectSphere.projectSphereMM(  numpy.array([xyzMM[sphere]]),
                                                                                    numpy.array([radiiMM[sphere]]),
                                                                                    sourceDetectorDistMM=sourceDetectorDistMM,
                                                                                    pixelSizeMM=pixelSizeMM,
                                                                                    detectorResolution=detectorResolution,
                                                                                    ROIcentreMM=xyzMM[sphere].copy(),
                                                                                    ROIradiusMM=radiiMM[sphere]*1.2,
                                                                                    transformationCentreMM=transformationCentresMM[n],
                                                                                    transformationMatrix=transformationMatrices[n])

                # Pre-allocate sens field, since it's local it should be the same shape as sphereRefProjectionMM
                sensXYZ = numpy.zeros((3, sphereRefProjectionMM.shape[0], sphereRefProjectionMM.shape[1]), dtype=float)

                # For each direction in X, Y, Z
                for i in range(3):
                    # Perturb just one direction
                    tmp = xyzMM[sphere].copy()
                    tmp[i] += perturbationMM[i]
                    # Here the ROI parameters are the same as the undisturbed sphere to guarantee the same size
                    #   projection being returned
                    perturbedProjectionMM = radioSphere.projectSphere.projectSphereMM(  numpy.array([tmp]),
                                                                                        numpy.array([radiiMM[sphere]]),
                                                                                        sourceDetectorDistMM=sourceDetectorDistMM,
                                                                                        pixelSizeMM=pixelSizeMM,
                                                                                        detectorResolution=detectorResolution,
                                                                                        ROIcentreMM=xyzMM[sphere].copy(),
                                                                                        ROIradiusMM=radiiMM[sphere]*1.2,
                                                                                        transformationCentreMM=transformationCentresMM[n],
                                                                                        transformationMatrix=transformationMatrices[n])
                    sensXYZ[i] = sphereRefProjectionMM - perturbedProjectionMM

                # Append sens field and residual ROI to lists for optimisation
                sensXYZall.append(sensXYZ)
                residualMMroi.append(residualMM[n, limits[0,0]:limits[0,1], limits[1,0]:limits[1,1]])

            if GRAPH:
                for n in range(nProj):
                    for i in range(3):
                        plt.subplot(nProj, 5, n*5+3+i)
                        plt.title("Proj {} Sens{} (mm) LUT: [-2, 2]\nSphere = {}, Perturbation {} mm".format(n, "XYZ"[i], sphere, perturbationMM[i]))
                        plt.imshow(sensXYZall[n][i], cmap='coolwarm', vmin=-2, vmax=2)
                plt.show()

            # Reset step which will be filled with the step in XZY combining weighted combination of sensitivity fields
            #   N.B. not to be confused with perturbationMM, which for now is not changed during iterations
            step = numpy.zeros(3)
            if solver == 'homemade':
                        #mask = numpy.where(numpy.abs(residualMM) > 0.01)
                        #for i in range(3):
                            #tmp = residualMM / sensXYZ[i]

                            #step[i] = -1.0 * numpy.mean(tmp[mask][numpy.isfinite(tmp[mask])])
                            #xyzMM[sphere][i] += step[i] * perturbationMM[i]
                            #if GRAPH:
                                #plt.subplot(2,4,4+i+2)
                                #plt.imshow(tmp, cmap='coolwarm', vmin=-5, vmax=5)
                                #plt.title("Step = {}".format(step[i]))
                        #if verbose: print("step:\t",step)
                        #if verbose: print("pos:\t",xyzMM[sphere])
                        pass

            elif solver == 'leastsquares':
                        # attempt to find weights a, b, c for the x, y, z sensitivity fields with leastsquares
                        import scipy.optimize
                        def optimiseMe(weights, residualMMroi, sensXYZall):
                            #print("optimiseMe(): weights = ", weights)
                            assert(len(residualMMroi) == len(sensXYZall))

                            # Add a series of 1D errors for each radio
                            output = []
                            for n in range(len(residualMMroi)):
                                    output = numpy.hstack([output, numpy.ravel(residualMMroi[n] - weights[0]*sensXYZall[n][0]
                                                                                                - weights[1]*sensXYZall[n][1]
                                                                                                - weights[2]*sensXYZall[n][2])])
                            return output

                        LSQret = scipy.optimize.least_squares(  optimiseMe,
                                                                [1.0, 1.0, 1.0], # this is step, i.e., the *weights* for the different perturbations, they are not in MM
                                                                args=[residualMMroi, sensXYZall],
                                                                verbose=False,
                                                                method='lm',
                                                                diff_step=1.0)
                        if LSQret['success'] == True:
                            #print('LSQ Success!')
                            step = LSQret['x']
                            for i in range(3):
                                xyzMM[sphere][i] -= step[i] * perturbationMM[i]
                        else:
                            print("LSQ failed to converge")

                        if verbose > 1:
                            print("\t\tstep:\t", step)
                            print("\t\tpos:\t",  xyzMM[sphere])

            else:
                print("optimiseSensitivityFieldsMultiProj(): Don't know this solver")
                return

        ### Now a soothing  DEM step can be applied to the current updated XYZ positions
        if DEMcorr:
            import radioSphere.DEM
            xyzMMnew = radioSphere.DEM.DEM_step(xyzMM,radiiMM)
            if verbose > 0: print("   DEM changed positions by: {:0.3f}mm".format(numpy.linalg.norm(xyzMM-xyzMMnew)), end='')
            xyzMM = xyzMMnew

        if GRAPH: plt.show()

        #if verbose > 0: print("   |deltaMM|: ", numpy.linalg.norm(xyzMM-xyzMMprev))
        if verbose > 0: print("   |deltaMM|: {:0.3f}".format(numpy.linalg.norm(step*perturbationMM)))

        xyzMMprev = xyzMM.copy()

        iterations += 1


    if verbose > 0:
        if numpy.linalg.norm(step*perturbationMM) <= minDeltaMM:
            # Check that we exited based on displacement
            print("\toptimiseSensitivityFieldsMultiProj(): Got below {} in {} iterations".format(minDeltaMM, iterations))
        else:
            # Check that we exited based on displacement
            print("\toptimiseSensitivityFieldsMultiProj(): hit max iterations ({})".format(iterations))

    return xyzMM



__all__ = ["optimisePositions", "projectSphere", "DEM", "detectSpheres", "calibrateAttenuation"]

#from .blender_viewer import *
from .optimisePositions import *
from .detectSpheres import *
from .projectSphere import *
from .DEM import DEM, mercury, nddem
from .calibrateAttenuation import *



'''
Implementation of tomopack by Stéphane Roux
'''

import os
import numpy
import radioSphere.projectSphere
import scipy.ndimage
from scipy.spatial import distance

import matplotlib.pyplot as plt
import tifffile

# _kernel = numpy.ones((1,3,3))/9. # square kernel
_kernel = numpy.array([[[1,2,1],[2,4,2],[1,2,1]]])/16. # gaussian kernel

# non-naiive approach
def _PI_projector(f):
    g = f.copy()
    g[f<0.5] = 0.
    # g[g.imag < 1e-14] = numpy.round(g[g.imag < 1e-14])
    # g.real = numpy.round(g.real)
    g = numpy.round(numpy.abs(g))
    return g

# Bottom of page 8 in TomoPack.pdf --- just take peaks in a 3x3 pixel area, and weight by all of the mass in that area
# Why not directly look at maxima in the masses, this avoids the
def _filter_maxima(f, debug=False, removeNegatives=False):
    # Detect unpadded 2D image first:
    if len(f.shape) == 2:
        f = f[numpy.newaxis, ...]
        twoD = True
    else:
        twoD = False

    # This is making our complex result into a scalar result
    f_abs = numpy.abs(f)

    # Remove negatives -- however keeping negatives can help sharpen the result of the convolution
    if removeNegatives: f_abs[f<0] = 0

    masses = scipy.ndimage.convolve(f_abs,_kernel)

    peaks = masses == scipy.ndimage.maximum_filter(masses,size=(1,3,3))

    if debug:
        #plt.imshow(numpy.real(f))
        #plt.title('real part of original estimation')
        #plt.colorbar()
        #plt.show()

        #plt.imshow(f_abs)
        #plt.title('positive real component')
        #plt.colorbar()
        #plt.show()

        #plt.imshow(g)
        #plt.colorbar()
        #plt.title('Local filtered maxima')
        #plt.show()

        #plt.imshow(peaks)
        #plt.colorbar()
        #plt.title('Location of local peaks')
        #plt.show()

        plt.imshow(masses[0])
        plt.colorbar()
        plt.title('Local masses (first slice if 3D)')
        plt.show()

        plt.imshow(masses[0]>0.5)
        plt.colorbar()
        plt.title('Local masses > 0.5 (first slice if 3D)')
        plt.show()

        #plt.imshow(peaks*masses)
        #plt.colorbar()
        #plt.title('Final weighted estimation')
        #plt.show()

    if twoD: return peaks[0]*masses[0]
    else:    return peaks   *masses


def tomopack(radioMM, psiMM, maxIterations=50, l=0.1, epsilon='iterative', kTrustRatio=0.75, GRAPH=False):
    '''
    'tomopack' FFT-based algorithm for sphere identification by Stéphane Roux.
    the FFT of psiMM as a structuring element to pick up projected spheres in radioMM
    using a parallel projection hypothesis.

    Parameters
    ----------
        radioMM : 2D numpy array of floats
            radiography containing "distance" projections in mm
            This is typically the result of mu*log(I/I_0)

        psiMM : 2D numpy array of floats
            A (synthetic) projection of a single sphere in approx the same settings,
            as radioMM -- this is the structuring element used for FFT recognition.
            Should be same size as radioMM

        maxIterations : int (optional)
            Number of iterations to run the detection for.
            Default = 50

        l : float (optional)
            "lambda" parameter which controls relaxation in the iterations.
            Default = 0.1

        epsilon : float (optional), or 'iterative'
            trust cutoff in psi.
            If the radiograph is in mm so is this parameter
            Default = 'iterative'

        GRAPH : bool (optional)
            VERY noisily make graphs?

    Returns
    -------
        f_x : 2D numpy array
            same size as radioMM, high values where we think particles are.
            Consider using indicatorFunctionToDetectorPositions()

    '''
    #assert(psi.sum() < radioMM.sum()), "detectSpheres.tomopack(): sum(psi) > sum(radioMM), doesn't look right"
    assert(len(psiMM.shape) == 2), "detectSpheres.tomopack(): psi and radioMM should be same size"
    assert(len(psiMM.shape) == len(radioMM.shape)), "detectSpheres.tomopack(): psi and radioMM should be same size"
    assert(psiMM.shape[0] == radioMM.shape[0]), "detectSpheres.tomopack(): psi and radioMM should be same size"
    assert(psiMM.shape[1] == radioMM.shape[1]), "detectSpheres.tomopack(): psi and radioMM should be same size"

    radioMM_FFT = numpy.fft.fft2(radioMM)

    # define the psi function and its FFT -- the structuring element for SR
    # Start with the projection of a centred sphere
    #psiMM_FFT = numpy.fft.fft2(psiMM)
    # 2021-10-26 OS: phase shift of psi so that zero phase angle is positioned at the centre of the detector
    # avoids the HACK of shifting f_x later on line 187
    psiMM_FFT = numpy.fft.fft2(numpy.fft.fftshift(psiMM))


    # This is comparable to Figure 7 in TomoPack
    # if GRAPH:
    #     plt.imshow(1./numpy.abs(psiMM_FFT), vmin=0, vmax=1, cmap='hot')
    #     plt.title(r"$1/|\psi_{FFT}|$")
    #     plt.colorbar()
    #     plt.show()

    ## naiive approach
    with numpy.errstate(divide='ignore', invalid='ignore'):
        f_k_naiive = radioMM_FFT/psiMM_FFT
    f_x_naiive = numpy.fft.ifft2(f_k_naiive)


    # Prepare for iterations

    f_x = numpy.zeros_like(radioMM)
    if epsilon == 'iterative':
        epsilon = 1e-5
        k_trust = numpy.abs(psiMM_FFT) > epsilon
        while k_trust.sum()/k_trust.shape[0]/k_trust.shape[1] > kTrustRatio:
            # print(epsilon)
            epsilon *= 1.5
            k_trust = numpy.abs(psiMM_FFT) > epsilon
        f_x_old = epsilon*numpy.ones_like(radioMM)
    else:
        f_x_old = epsilon*numpy.ones_like(radioMM)
        k_trust = numpy.abs(psiMM_FFT) > epsilon

    if GRAPH: print("INFO: k_trust maintains {}% of wavenumubers".format(100*k_trust.sum()/k_trust.shape[0]/k_trust.shape[1]))

    count = 0
    if GRAPH: plt.ion()


    # Start iterations as per Algorithm 1 in TomoPack
    #   Objective: get a good indictor function f_x
    while numpy.linalg.norm(f_x - f_x_old) > 1e-8: # NOTE: Using a different epsilon here
        if GRAPH: plt.clf()

        f_x_old = f_x.copy()
        f_k = numpy.fft.fft2(f_x)
        # if GRAPH: plt.plot(f_k,'k.')
        f_k[k_trust] = f_k_naiive[k_trust]
        # if GRAPH: plt.plot(f_k,'r.')
        f_x = numpy.fft.ifft2(f_k)
        #f_x = numpy.fft.fftshift(numpy.fft.ifft2(f_k)) # importing HACK from 1D version
        f_x = f_x + l*(_PI_projector(f_x) - f_x)
        #print(numpy.amax(numpy.abs(f_x - f_x_old)))

        if GRAPH:
            plt.imshow(numpy.abs(f_x), vmin=0)#, vmax=1)
            plt.colorbar()
            # plt.plot(numpy.real(f_x),numpy.imag(f_x),'.')
            plt.pause(0.00001)
        count += 1

        if count > maxIterations:
            # f_x = numpy.zeros_like(f_x) # NOTE: THIS WILL RETURN NO MATCHES WHEN TOMOPACK DOESN'T CONVERGE
            print('\nKILLED LOOP')
            break
    # plt.show()
    # print('Took ' + str(count) + ' iterations to complete.')

    if GRAPH:
        #not_k_trust_ind = numpy.where(~k_trust)

        plt.figure(figsize=[8,6])

        plt.subplot(331)
        plt.title(r'$p(x)$')
        plt.imshow(radioMM)
        # plt.plot(pPredictedIJ[:,0], pPredictedIJ[:,1], 'rx')
        plt.colorbar()

        plt.subplot(332)
        plt.title(r'$\psi(x)$')
        plt.imshow(psiMM, vmin=radioMM.min(), vmax=radioMM.max())
        plt.colorbar()

        ax3 = plt.subplot(333)
        plt.title(r'True $f(x)$')
        # HACK ky-flat image
        plt.imshow(numpy.zeros_like(radioMM))
        #ax3.set_aspect('equal', 'datalim')

        plt.subplot(334)
        plt.title(r'$|\tilde{p}(k)|$')
        plt.imshow(numpy.abs(radioMM_FFT))#, vmin=0, vmax=10)
        plt.colorbar()

        plt.subplot(335)
        plt.title(r'$|\tilde{\psi}(k)|$')
        plt.imshow(numpy.abs(psiMM_FFT), vmin=0, vmax=10)
        plt.colorbar()

        one_on_psi_plot = 1./numpy.abs(psiMM_FFT)
        one_on_psi_plot[~k_trust] = numpy.nan
        plt.subplot(336)
        plt.title(r'$1/|\psi(k)|$')
        plt.imshow(one_on_psi_plot)
        # plt.semilogy(numpy.arange(len(psi_FFT))[k_trust],1./numpy.abs(psi_FFT)[k_trust],'g.')
        #plt.semilogy(numpy.arange(len(psi_FFT))[~k_trust],1./numpy.abs(psi_FFT)[~k_trust],'r.')
        #plt.semilogy([0,len(psi_FFT)],[1./epsilon,1./epsilon],'k--')
        # plt.plot(not_k_trust_ind[0], not_k_trust_ind[1], 'r.')
        plt.colorbar()
        # plt.ylim(0,1./epsilon)


        plt.subplot(337)
        plt.title(r'Estimated $|f(k)|$')
        plt.imshow(numpy.abs(f_k))
        plt.colorbar()

        plt.subplot(338)
        plt.title(r'Estimated $f(x)$')
        plt.imshow(numpy.real(f_x))
        # plt.plot(pPredictedIJ[:,0], pPredictedIJ[:,1], 'o', mec='r', mfc='None', mew=0.1)
        plt.colorbar()

        #plt.subplot(339)
        #plt.title(r'Residual $p(x) - p(f(x))$')
        #p_f_x = numpy.zeros_like(radioMM, dtype=float)


        ### Convert to XYZ in space and mm
        #positionsXYZmm = numpy.zeros([pPredictedIJ.shape[0], 3])
        ## Y
        #positionsXYZmm[:,1] = -1*(pPredictedIJ[:,0] - radioMM.shape[0]/2.0)*pixelSizeMM
        ## Z
        #positionsXYZmm[:,2] = -1*(pPredictedIJ[:,1] - radioMM.shape[1]/2.0)*pixelSizeMM
        #radii = numpy.ones([pPredictedIJ.shape[0]])
        #radii *= radiusMM

        #print("\n\n\n", pPredictedIJ, "\n\n\n")
        #print("\n\n\n", positionsXYZmm, "\n\n\n")

        #p_f_x = radioSphere.projectSphere.projectSphereMM(positionsXYZmm,
                                                          #radii,
                                                          #sourceDetectorDistMM=numpy.inf,
                                                          #pixelSizeMM=pixelSizeMM,
                                                          #detectorResolution=detectorResolution)

        #residual = p_f_x
        #residual = p_f_x - radioMM
        #vmax = numpy.abs(residual).max()
        #plt.imshow(residual, vmin=-vmax, vmax=vmax, cmap='coolwarm')
        #plt.colorbar()



        plt.subplots_adjust(hspace=0.5)
        plt.show()
        #plt.savefig('tt3D.png',dpi=200)

    #return pPredictedIJ
    return numpy.real(f_x)


def indicatorFunctionToDetectorPositions(f_x, debug=False, particlesPresenceThreshold=0.9):
    '''
    Takes an approximation of the indicator function f_x from tomopack and returns positions on the detector

    Parameters
    ----------
        f_x : 2D numpy array of floats
            The value of the indicator function for each pixel on the detector

        debug : bool, optional
            Show debug graphs (especially in the maximum filtering)
            Default = False

        particlesPresence : float, optional
            Threshold for accepting a particle
            Default = 0.9
    '''
    assert(len(f_x.shape) == 2), "detectSpheres.indicatorFunctionToDetectorPositions(): need 2D array"

    f_x = _filter_maxima(f_x, debug=debug)

    particlesPresence = numpy.round(numpy.real(_PI_projector(f_x)))
    pPredictedIJ = []
    for pos in numpy.array(numpy.where(particlesPresence > particlesPresenceThreshold)).T:
        val = particlesPresence[pos[0], pos[1]]
        # print(f_x[pos[0],pos[1]])
        for i in range(int(val)):
            pPredictedIJ.append([pos[1], pos[0]])
    pPredictedIJ = numpy.array(pPredictedIJ)

    return pPredictedIJ


def psiSeriesScanTo3DPositions(radio,      # not (necessarily) MM
                               psiXseries, # Obviously in the same units as radio above, please
                               radiusMM,   # can be removed?
                               CORxPositions = None,
                               massThreshold=0.12,
                               scanPersistenceThresholdRadii=None,
                               scanFixedNumber=None,
                               scanPersistenceThreshold=7, maxIterations=50,
                               sourceDetectorDistMM=100, pixelSizeMM=0.1, l=0.2, kTrustRatio=0.7, useCache=True,
                               numCores=1,
                               blur=0.0,
                               cacheFile='fXseries.tif',
                               verbose=False):

    # it is our objective to fill in fx series
    fXseries = numpy.zeros_like(psiXseries)

    if CORxPositions is None:
        print("xPositions is not passed, just putting 1, 2, 3...")
        CORxPositions = numpy.arange(psiXseries.shape[0])

    for posN, CORxPos in enumerate(CORxPositions):
        ### "Structuring Element"
        print("\t{}/{} CORxPos = {:0.2f} mm".format(posN+1, len(CORxPositions), CORxPos), end='\r')
        fXseries[posN] = radioSphere.detectSpheres.tomopack(radio,
                                                            psiXseries[posN],
                                                            GRAPH=0,
                                                            maxIterations=maxIterations,
                                                            l=l,
                                                            kTrustRatio=kTrustRatio)
    tifffile.imsave(cacheFile, fXseries.astype('float'))
    print(f"saved {cacheFile}")
    #loadedCache = False

    #if useCache:
        #cachePsiFile = cacheFile[:-4] + '_psi.tif'
        #if os.path.isfile(cacheFile) and os.path.isfile(cachePsiFile):
            #print("Loading previous indicator functions... ", end="")
            #fXseries = tifffile.imread(cacheFile)
            #psiXseries = tifffile.imread(cachePsiFile)
            #if ( fXseries.shape[0] == CORxNumber ) and ( fXseries.shape[1] == radioMM.shape[0] ) and ( fXseries.shape[2] == radioMM.shape[1] ):
                #print("done.")
                #loadedCache = True
            #else:
                #print("cached file had wrong dimensions. Generating new cache file.")
        #else:
            #print('No cached indicator functions found. Generating them now to cache.')
    #if not loadedCache:
        #fXseries = numpy.zeros((len(CORxPositions), radioMM.shape[0], radioMM.shape[1]))
        #psiXseries = numpy.zeros_like(fXseries)

        #psiRefMM = radioSphere.projectSphere.projectSphereMM(numpy.array([[(CORxMax+CORxMin)/2., 0., 0.]]),
                                                          #numpy.array([radiusMM]),
                                                          #detectorResolution=radioMM.shape,
                                                          #pixelSizeMM=pixelSizeMM,
                                                          #sourceDetectorDistMM=sourceDetectorDistMM,
                                                          #blur=blur)

        #for posN, CORxPos in enumerate(CORxPositions):
            #### "Structuring Element"
            #print("\t{}/{} CORxPos = {:0.2f}mm".format(posN+1, len(CORxPositions), CORxPos), end='\r')
            #psiMM = radioSphere.projectSphere.projectSphereMM(numpy.array([[CORxPos, 0., 0.]]),
                                                              #numpy.array([radiusMM]),
                                                              #detectorResolution=radioMM.shape,
                                                              #pixelSizeMM=pixelSizeMM,
                                                              #sourceDetectorDistMM=sourceDetectorDistMM,
                                                              #blur=blur)

            #fXseries[posN] = radioSphere.detectSpheres.tomopack(radioMM, psiMM, GRAPH=0, maxIterations=maxIterations, l=l, kTrustRatio=kTrustRatio)
            #psiXseries[posN] = radioSphere.detectSpheres.tomopack(psiRefMM, psiMM, GRAPH=0, maxIterations=maxIterations, l=l, kTrustRatio=kTrustRatio)

    #if useCache and not loadedCache:
        #print("Saving indicator functions for next time... ", end="")
        #tifffile.imsave(cacheFile, fXseries.astype('<f4'))
        #tifffile.imsave(cachePsiFile, psiXseries.astype('<f4'))
        #print("done.")

    #L_x  = 20 # TODO: SCALING IN X DIRECTION SHOULD BE A FUNCTION OF THE CONE ANGLE
    #L_yz =  2 # TODO: THIS SHOULD BE A FUNCTION OF THE PIXELS PER RADIUS

    #struct = psiXseries[(psiXseries.shape[0])//2 -  L_x:(psiXseries.shape[0])//2 + L_x  + 1,
                        #(psiXseries.shape[1])//2 - L_yz:(psiXseries.shape[1])//2 + L_yz + 1,
                        #(psiXseries.shape[2])//2 - L_yz:(psiXseries.shape[2])//2 + L_yz + 1]

    #fXconvolvedSeries = scipy.ndimage.convolve(fXseries,struct/struct.sum())
    ##if useCache and not loadedCache:
        ##tifffile.imsave(f'{cacheFile[:-4]}_struct.tif', struct.astype('<f4'))
        ##tifffile.imsave(f'{cacheFile[:-4]}_fXconvolvedSeries.tif', fXconvolvedSeries.astype('<f4'))


    #binaryPeaks = fXconvolvedSeries > massThreshold

    zoomLevel = sourceDetectorDistMM/((CORxPositions[0] + CORxPositions[-1])/2)
    CORxDelta = numpy.abs(CORxPositions[0]-CORxPositions[1])
    # Look in a volume of +/- half a radius in all directions for the highest value (+/- 1 radius keeps overlapping and causing issues, half a radius doesn't overlap particles, but still contains one clean peak)
    fXseriesMaximumFiltered = scipy.ndimage.maximum_filter(fXseries,
                                                              size=(3*numpy.int(numpy.floor(radiusMM/CORxDelta)),
                                                                    numpy.int(numpy.floor(radiusMM/pixelSizeMM*zoomLevel)),
                                                                    numpy.int(numpy.floor(radiusMM/pixelSizeMM*zoomLevel)))
                                                              )
    allPeaks = fXseries == fXseriesMaximumFiltered
    masses = allPeaks*fXseries

    if verbose:
        tifffile.imsave(cacheFile[:-4] + '_masses.tif', masses.astype('<f4'))
        tifffile.imsave(cacheFile[:-4] + '_peaks.tif', allPeaks.astype('<f4'))
        tifffile.imsave(cacheFile[:-4] + '_fXseries.tif', fXseries.astype('<f4'))
        tifffile.imsave(cacheFile[:-4] + '_fXseriesMaximumFiltered.tif', fXseriesMaximumFiltered.astype('<f4'))

    if scanFixedNumber:
        # get the indices of all of the peaks, from highest to lowest
        sortedPeakIndices = numpy.argsort(masses, axis=None)[::-1]
        # print(sortedPeakIndices.shape)
        # get just the first scanFixedNumber of those and put them into a scanFixedNumber x 3 array
        peaksCORxPOSnJI = numpy.vstack(numpy.unravel_index(sortedPeakIndices[:scanFixedNumber], masses.shape)).T
        # print(peaksCORxPOSnJI.shape)
    else:
        filteredPeaks = masses > massThreshold
        peaksCORxPOSnJI = numpy.argwhere(filteredPeaks)

        if verbose: tifffile.imsave(cacheFile[:-4] + '_filteredPeaks.tif', filteredPeaks.astype('<f4'))
    # print(peaksCORxPOSnJI)

    ###############################################################
    ### Now we have guesses for all particle according to detector
    ###   (IJ) and position along the X-scanning direction
    ### We're going to convert that to spatial XYZ
    ###############################################################
    print("\nConverting tomopack x-scan to 3D positions\n")
    ## Convert to XYZ in space and mm
    positionsXYZmm = numpy.zeros([peaksCORxPOSnJI.shape[0], 3])

    for i in range(positionsXYZmm.shape[0]):
        # X -- look up which CORx slice the maximum falls in, this could be interpolated instead of rounded
        positionsXYZmm[i,0] = CORxPositions[int(numpy.round(peaksCORxPOSnJI[i,0]))]

        # detector I gives real position Y in mm
        yPosDetMM = -1*(peaksCORxPOSnJI[i,2] - radio.shape[1]/2.0)*pixelSizeMM

        # detector J gives real position Z in mm
        zPosDetMM = -1*(peaksCORxPOSnJI[i,1] - radio.shape[0]/2.0)*pixelSizeMM

        # And now scale down by zoom factor
        # Y
        positionsXYZmm[i,1] = yPosDetMM * ( positionsXYZmm[i,0] / sourceDetectorDistMM )
        # Z
        positionsXYZmm[i,2] = zPosDetMM * ( positionsXYZmm[i,0] / sourceDetectorDistMM )

    print(f"\ntomopackDivergentScanTo3DPositions(): I'm returning {positionsXYZmm.shape[0]} 3D positions.\n")
    return positionsXYZmm


def tomopackDivergentScanTo3DPositions(radioMM, radiusMM,
                                       CORxMin=None, CORxMax=None, CORxNumber=100,
                                       psiXseries = None,
                                       massThreshold=0.12,
                                       scanPersistenceThresholdRadii=None,
                                       scanFixedNumber=None,
                                       scanPersistenceThreshold=7, maxIterations=50,
                                       sourceDetectorDistMM=100, pixelSizeMM=0.1, l=0.2, kTrustRatio=0.7, useCache=True,
                                       numCores=1,
                                       blur=0.0,
                                       cacheFile='fXseries.tif',
                                       verbose=False):
    """
    This function takes in a single divergent projection, and will run tomopack
    generating different psis by varying their position as Centre Of Rotation (COR) in the x-direction,
    from CORxMin to CORxMax in CORxNumber steps.

    The resulting series of indicator functions is analysed and a 3D position guess for
    all identified spheres is returned.

    Parameters
    ----------
        massThreshold : float (optional)
            Threshold for the accepting the result of the convolution of the indicator function.
            Deafult = 0.5

        blur : float, optional
            sigma of blur to pass to scipy.ndimage.gaussian_filter to
            blur the radiograph at the end of everything

        # scanPersistenceThresholdRadii : float (optional)
            # How much +/- in radii in the scanning direction should the particle's indicator function be above massThreshold? If not passed, scanPersistenceThreshold is used

    Returns
    -------
        positionsXYZmm : 2D numpy array
    """
    assert(len(radioMM.shape) == 2), "detectSpheres.tomopackDivergentScanTo3DPositions(): need 2D array"

    if CORxMin is None: CORxMin = CORx-3*radiusMM
    if CORxMax is None: CORxMax = CORx+3*radiusMM

    CORxPositions = numpy.linspace(CORxMin, CORxMax, CORxNumber)
    CORxDelta = CORxPositions[1] - CORxPositions[0]

    loadedCache = False

    if useCache:
        cachePsiFile = cacheFile[:-4] + '_psi.tif'
        if os.path.isfile(cacheFile) and os.path.isfile(cachePsiFile):
            print("Loading previous indicator functions... ", end="")
            fXseries = tifffile.imread(cacheFile)
            psiXseries = tifffile.imread(cachePsiFile)
            if ( fXseries.shape[0] == CORxNumber ) and ( fXseries.shape[1] == radioMM.shape[0] ) and ( fXseries.shape[2] == radioMM.shape[1] ):
                print("done.")
                loadedCache = True
            else:
                print("cached file had wrong dimensions. Generating new cache file.")
        else:
            print('No cached indicator functions found. Generating them now to cache.')
    if not loadedCache:
        fXseries = numpy.zeros((len(CORxPositions), radioMM.shape[0], radioMM.shape[1]))
        psiXseries = numpy.zeros_like(fXseries)
        psiMMseries = numpy.zeros_like(fXseries)

        psiRefMM = radioSphere.projectSphere.projectSphereMM(numpy.array([[(CORxMax+CORxMin)/2., 0., 0.]]),
                                                          numpy.array([radiusMM]),
                                                          detectorResolution=radioMM.shape,
                                                          pixelSizeMM=pixelSizeMM,
                                                          sourceDetectorDistMM=sourceDetectorDistMM,
                                                          blur=blur)

        for posN, CORxPos in enumerate(CORxPositions):
            ### "Structuring Element"
            print("\t{}/{} CORxPos = {:0.2f}mm".format(posN+1, len(CORxPositions), CORxPos), end='\r')
            psiMM = radioSphere.projectSphere.projectSphereMM(numpy.array([[CORxPos, 0., 0.]]),
                                                              numpy.array([radiusMM]),
                                                              detectorResolution=radioMM.shape,
                                                              pixelSizeMM=pixelSizeMM,
                                                              sourceDetectorDistMM=sourceDetectorDistMM,
                                                              blur=blur)

            fXseries[posN] = radioSphere.detectSpheres.tomopack(radioMM, psiMM, GRAPH=0, maxIterations=maxIterations, l=l, kTrustRatio=kTrustRatio)
            psiXseries[posN] = radioSphere.detectSpheres.tomopack(psiRefMM, psiMM, GRAPH=0, maxIterations=maxIterations, l=l, kTrustRatio=kTrustRatio)
            psiMMseries[posN] = psiMM

    if useCache and not loadedCache:
        print("Saving indicator functions for next time... ", end="")
        tifffile.imsave(cacheFile, fXseries.astype('<f4'))
        tifffile.imsave(cachePsiFile, psiXseries.astype('<f4'))
        tifffile.imsave(cachePsiFile[:-4]+'_MM.tif', psiMMseries.astype('<f4'))
        print("done.")


    use_old_method = False
    if use_old_method:
        ### Step 1. Maximum filtering... Trying only in the detector direction, in order not to bridge over the bad
        #           COR x positions where there are weird resonances
        ### Step 2. Lump together objects in the space of IJ and X scan
        binaryPeaks = scipy.ndimage.convolve(fXseries,_kernel) > massThreshold
        structure = numpy.ones([5,3,3])
        binaryPeaks = scipy.ndimage.binary_closing(binaryPeaks,structure=structure) # remove any gaps
        if verbose: tifffile.imsave('closed_peaks.tif', binaryPeaks.astype('<f4'))
    else:
        L_x  = 20 # TODO: SCALING IN X DIRECTION SHOULD BE A FUNCTION OF THE CONE ANGLE
        L_yz =  2 # TODO: THIS SHOULD BE A FUNCTION OF THE PIXELS PER RADIUS

        struct = psiXseries[(psiXseries.shape[0])//2 -  L_x:(psiXseries.shape[0])//2 + L_x  + 1,
                            (psiXseries.shape[1])//2 - L_yz:(psiXseries.shape[1])//2 + L_yz + 1,
                            (psiXseries.shape[2])//2 - L_yz:(psiXseries.shape[2])//2 + L_yz + 1]

        fXconvolvedSeries = scipy.ndimage.convolve(fXseries,struct/struct.sum())
        if useCache and not loadedCache:
            tifffile.imsave(f'{cacheFile[:-4]}_struct.tif', struct.astype('<f4'))
            tifffile.imsave(f'{cacheFile[:-4]}_fXconvolvedSeries.tif', fXconvolvedSeries.astype('<f4'))


        binaryPeaks = fXconvolvedSeries > massThreshold

    get_peaks_from_persistence = False
    if get_peaks_from_persistence:

        labelledPeaks = scipy.ndimage.label(binaryPeaks)[0]
        labelledPeaksBB = scipy.ndimage.find_objects(labelledPeaks)

        print(f"Found {labelledPeaks.max()} peaks across the indicator function series.")
        # Compute centres of mass and persistence in the scanning direction
        peaksCORxPOSnJI = numpy.zeros((labelledPeaks.max(), 4))

        try:
            import spam.label
            #print("Doing it with spam")
            BB = spam.label.boundingBoxes(labelledPeaks)

            # save x-y-z positions where x is the slice number in the scan direction and y, z pixel positions on detector
            peaksCORxPOSnJI[:,0:3] = spam.label.centresOfMass(labelledPeaks, boundingBoxes=BB)[1:]

            # also add in how many slices this is alive for
            peaksCORxPOSnJI[:,-1] = BB[1:,1] - BB[1:,0] + 1
            #print("Done it with spam")

        except:
            print("Not doing peak label analysis with spam (not found) this will be very slow")
            for label in range(1,labelledPeaks.max()+1):
                # save x-y-z positions where x is the slice number in the scan direction and y, z pixel positions on detector
                peaksCORxPOSnJI[label-1][0:3] = scipy.ndimage.center_of_mass(labelledPeaks==label)

                # also add in how many slices this is alive for
                peaksCORxPOSnJI[label-1][-1] = labelledPeaksBB[label-1][0].stop - labelledPeaksBB[label-1][0].start + 1
            print("...it was slow")

        if verbose: print("tomopackDivergentScanTo3DPositions(): object persistence in scan direction:\n", peaksCORxPOSnJI[:, -1])

        # Apply persistence threshold:
        persistenceMask = numpy.zeros(labelledPeaks.max(), dtype=bool)
        CORxSpacing = CORxPositions[1] - CORxPositions[0]
        if scanPersistenceThresholdRadii is not None:
            scanPersistenceThreshold = int(numpy.floor((scanPersistenceThresholdRadii*radiusMM)/CORxSpacing))
            print(f"Calculated scan persistence threshold for {scanPersistenceThresholdRadii}\% particle radii: {scanPersistenceThreshold}")
        if scanFixedNumber:
            # print('Scanning for a fixed number of particles is not implemented!')
            sortedPeakIndices = numpy.argsort(peaksCORxPOSnJI[:, -1])[::-1]
            for i in range(scanFixedNumber):
                persistenceMask[sortedPeakIndices[i]] = True
        else: # use threshold value
            persistenceMask[peaksCORxPOSnJI[:, -1] > scanPersistenceThreshold] = True
        # Apply mask and remove 4th column:
        peaksCORxPOSnJI = peaksCORxPOSnJI[persistenceMask, 0:3]
    else:
        zoomLevel = sourceDetectorDistMM/((CORxMin + CORxMax)/2)
        # Look in a volume of +/- half a radius in all directions for the highest value (+/- 1 radius keeps overlapping and causing issues, half a radius doesn't overlap particles, but still contains one clean peak)
        fXconvolvedMaximumFiltered = scipy.ndimage.maximum_filter(fXconvolvedSeries,
                                                                  size=(numpy.int(numpy.floor(radiusMM/CORxDelta)),
                                                                        numpy.int(numpy.floor(radiusMM/pixelSizeMM*zoomLevel)),
                                                                        numpy.int(numpy.floor(radiusMM/pixelSizeMM*zoomLevel))
                                                                        )
                                                                  )
        allPeaks = fXconvolvedSeries == fXconvolvedMaximumFiltered
        masses = allPeaks*fXconvolvedSeries
        if verbose:
            tifffile.imsave(cacheFile[:-4] + '_masses.tif', masses.astype('<f4'))
            tifffile.imsave(cacheFile[:-4] + '_peaks.tif', allPeaks.astype('<f4'))
            tifffile.imsave(cacheFile[:-4] + '_fXconvolvedSeries.tif', fXconvolvedSeries.astype('<f4'))
            tifffile.imsave(cacheFile[:-4] + '_fXconvolvedMaximumFiltered.tif', fXconvolvedMaximumFiltered.astype('<f4'))

        if scanFixedNumber:
            # get the indices of all of the peaks, from highest to lowest
            sortedPeakIndices = numpy.argsort(masses, axis=None)[::-1]
            # print(sortedPeakIndices.shape)
            # get just the first scanFixedNumber of those and put them into a scanFixedNumber x 3 array
            peaksCORxPOSnJI = numpy.vstack(numpy.unravel_index(sortedPeakIndices[:scanFixedNumber], masses.shape)).T
            # print(peaksCORxPOSnJI.shape)
        else:
            filteredPeaks = masses > massThreshold
            peaksCORxPOSnJI = numpy.argwhere(filteredPeaks)

            if verbose: tifffile.imsave(cacheFile[:-4] + '_filteredPeaks.tif', filteredPeaks.astype('<f4'))
        # print(peaksCORxPOSnJI)




    ###############################################################
    ### Now we have guesses for all particle according to detector
    ###   (IJ) and position along the X-scanning direction
    ### We're going to convert that to spatial XYZ
    ###############################################################
    print("\nConverting tomopack x-scan to 3D positions\n")
    ## Convert to XYZ in space and mm
    positionsXYZmm = numpy.zeros([peaksCORxPOSnJI.shape[0], 3])

    for i in range(positionsXYZmm.shape[0]):
        # X -- look up which CORx slice the maximum falls in, this could be interpolated instead of rounded
        positionsXYZmm[i,0] = CORxPositions[int(numpy.round(peaksCORxPOSnJI[i,0]))]

        # detector I gives real position Y in mm
        yPosDetMM = -1*(peaksCORxPOSnJI[i,2] - radioMM.shape[1]/2.0)*pixelSizeMM

        # detector J gives real position Z in mm
        zPosDetMM = -1*(peaksCORxPOSnJI[i,1] - radioMM.shape[0]/2.0)*pixelSizeMM

        # And now scale down by zoom factor
        # Y
        positionsXYZmm[i,1] = yPosDetMM * ( positionsXYZmm[i,0] / sourceDetectorDistMM )
        # Z
        positionsXYZmm[i,2] = zPosDetMM * ( positionsXYZmm[i,0] / sourceDetectorDistMM )

    print(f"\ntomopackDivergentScanTo3DPositions(): I'm returning {positionsXYZmm.shape[0]} 3D positions.\n")
    return positionsXYZmm

def removeParticle(positionsXYZmm,residual,radioMM,radiiMM,pixelSizeDetectorMM,zoomLevel,sourceObjectDistMM,verbose,GRAPH):
    if verbose: print('Removing a particle')
    # find the location of the highest peak on the detector panel. This is presumably the centroid.
    residualMMPeakIndices = numpy.unravel_index(numpy.argmax(residual, axis=None), residual.shape)
    if verbose: print(f'residualMMPeakIndices: {residualMMPeakIndices}')

    # define unit vector between source and peak location
    yPosDetMM = -1*(residualMMPeakIndices[1] - radioMM.shape[1]/2.0)*pixelSizeDetectorMM
    zPosDetMM = -1*(residualMMPeakIndices[0] - radioMM.shape[0]/2.0)*pixelSizeDetectorMM
    magnitude = numpy.sqrt(zoomLevel**2*sourceObjectDistMM**2 + yPosDetMM**2 + zPosDetMM**2)
    s = numpy.array([
        zoomLevel*sourceObjectDistMM/magnitude,
        yPosDetMM/magnitude,
        zPosDetMM/magnitude])
    if verbose: print(f's: {s}')

    # find distance of every particle from the line defined by this unit vector
    distances = numpy.linalg.norm(numpy.cross(positionsXYZmm,s),axis=1)
    if verbose: print(f'distances: {distances}')

    # remove the particle closest to the line
    closestParticleIndex = numpy.argmin(distances)
    positionsXYZmm = numpy.delete(positionsXYZmm, closestParticleIndex, axis=0)

    p_f_x = radioSphere.projectSphere.projectSphereMM(positionsXYZmm,
                                                      radiiMM[0]*numpy.ones(len(positionsXYZmm)),
                                                      sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                      pixelSizeMM=pixelSizeDetectorMM,
                                                      detectorResolution=radioMM.shape)
    residual = p_f_x - radioMM

    if GRAPH:
        plt.imshow(residual)
        plt.show()

    return positionsXYZmm, residual

def addParticle(*args, **kwargs):
    # return addParticleRaster(*args, **kwargs)
    return addParticleSensitivity(*args, **kwargs)


def addParticleRaster(positionsXYZmm,residual,radioMM,radiiMM,pixelSizeDetectorMM,zoomLevel,sourceObjectDistMM,CORxMin, CORxMax,CORxNumber,verbose,GRAPH):
    if verbose: print('Adding a particle')
    # find the location of the highest peak on the detector panel. This is presumably the centroid.
    residualMMPeakIndices = numpy.unravel_index(numpy.argmin(residual, axis=None), residual.shape)
    if verbose: print(f'residualMMPeakIndices: {residualMMPeakIndices}')

    # define unit vector between source and peak location
    yPosDetMM = -1*(residualMMPeakIndices[1] - radioMM.shape[1]/2.0)*pixelSizeDetectorMM
    zPosDetMM = -1*(residualMMPeakIndices[0] - radioMM.shape[0]/2.0)*pixelSizeDetectorMM
    magnitude = numpy.sqrt(zoomLevel**2*sourceObjectDistMM**2 + yPosDetMM**2 + zPosDetMM**2)
    s = numpy.array([
        zoomLevel*sourceObjectDistMM/magnitude,
        yPosDetMM/magnitude,
        zPosDetMM/magnitude])
    if verbose: print(f's: {s}')

    # trying to find an optimal solution by doing a raster scan and looking for minimal residual
    x_test = numpy.linspace(CORxMin,CORxMax,CORxNumber)
    best_index = 0
    best_residual = numpy.inf
    ref_projection = radioSphere.projectSphere.projectSphereMM(positionsXYZmm,
                                                      radiiMM[0]*numpy.ones(len(positionsXYZmm)),
                                                      sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                      pixelSizeMM=pixelSizeDetectorMM,
                                                      detectorResolution=radioMM.shape)

    limits = radioSphere.projectSphere.singleSphereToDetectorPixelRange(s*x_test[0],
                                                                        radiiMM[0],
                                                                        radiusMargin=0.1,
                                                                        sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                                        pixelSizeMM=pixelSizeDetectorMM,
                                                                        detectorResolution=radioMM.shape)

    ref_projection_crop = ref_projection[limits[0,0]:limits[0,1], limits[1,0]:limits[1,1]]
    radioMM_crop = radioMM[limits[0,0]:limits[0,1], limits[1,0]:limits[1,1]]
    for i,x in enumerate(x_test):
        single_particle_projection = radioSphere.projectSphere.projectSphereMM(numpy.expand_dims(s*x,axis=0),
                                                          numpy.expand_dims(radiiMM[0],axis=0),
                                                          sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                          pixelSizeMM=pixelSizeDetectorMM,
                                                          detectorResolution=radioMM.shape,
                                                          ROIcentreMM=s*x_test[0],
                                                          ROIradiusMM=radiiMM[0])

        residual = ref_projection_crop + single_particle_projection - radioMM_crop
        # print(i, (residual**2).sum(), best_residual, best_index)

        if (residual**2).sum() < best_residual:
            best_index = i
            best_residual = (residual**2).sum()

        # plt.ion()
        # plt.title(i)
        # plt.imshow(residual)
        # plt.pause(0.001)
    bestPositionXYZmm = s*x_test[best_index]

    print(f'Best location at {best_index}-th x value: {x_test[best_index]}')

    optimise = True
    if optimise:
        import radiosphere.optimsePositions
        bestPositionXYZmm = radioSphere.optimisePositions.optimiseSensitivityFields(radioMM,
                                                                                    numpy.expand_dims(bestPositionXYZmm,axis=0), # try the middle of the sample
                                                                                    numpy.expand_dims(radiiMM[0],axis=0),
                                                                                    perturbationMM=(0.01, 0.01, 0.01),
                                                                                    # perturbationMM=(0.5, 0.25, 0.25),
                                                                                    # perturbationMM=(1, 0.5, 0.5),
                                                                                    # perturbationMM=(3, 1, 1),
                                                                                    minDeltaMM=0.0001,
                                                                                    iterationsMax=500,
                                                                                    sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                                                    pixelSizeMM=pixelSizeDetectorMM,
                                                                                    detectorResolution=radioMM.shape,
                                                                                    verbose=False,
                                                                                    # GRAPH=True,
                                                                                    # NDDEM_output=True
                                                                                    )

    positionsXYZmm = numpy.vstack([positionsXYZmm, bestPositionXYZmm])
    p_f_x = radioSphere.projectSphere.projectSphereMM(positionsXYZmm,
                                                      radiiMM[0]*numpy.ones(len(positionsXYZmm)),
                                                      sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                      pixelSizeMM=pixelSizeDetectorMM,
                                                      detectorResolution=radioMM.shape)
    residual = p_f_x - radioMM

    if GRAPH:
        plt.imshow(residual)
        plt.show()

    return positionsXYZmm, residual


def addParticleSensitivity(positionsXYZmm,residual,radioMM,radiiMM,pixelSizeDetectorMM,zoomLevel,sourceObjectDistMM,CORxMin, CORxMax,CORxNumber,verbose,GRAPH):
    if verbose: print('Adding a particle')
    # find the location of the highest peak on the detector panel. This is presumably the centroid.
    residualMMPeakIndices = numpy.unravel_index(numpy.argmin(residual, axis=None), residual.shape)
    if verbose: print(f'residualMMPeakIndices: {residualMMPeakIndices}')

    # define unit vector between source and peak location
    yPosDetMM = -1*(residualMMPeakIndices[1] - radioMM.shape[1]/2.0)*pixelSizeDetectorMM
    zPosDetMM = -1*(residualMMPeakIndices[0] - radioMM.shape[0]/2.0)*pixelSizeDetectorMM
    magnitude = numpy.sqrt(zoomLevel**2*sourceObjectDistMM**2 + yPosDetMM**2 + zPosDetMM**2)
    s = numpy.array([
        zoomLevel*sourceObjectDistMM/magnitude,
        yPosDetMM/magnitude,
        zPosDetMM/magnitude])
    if verbose: print(f's: {s}')

    # trying to find an optimal solution by doing a raster scan and looking for minimal residual
    positionXYZmmOpt = radioSphere.optimisePositions.optimiseSensitivityFields(   radioMM,
                            numpy.expand_dims(s*sourceObjectDistMM,axis=0), # try the middle of the sample
                            numpy.expand_dims(radiiMM[0],axis=0),
                            perturbationMM=(0.001, 0.001, 0.001),
                            minDeltaMM=0.0001,
                            iterationsMax=500,
                            sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                            pixelSizeMM=pixelSizeDetectorMM,
                            detectorResolution=radioMM.shape,
                            verbose=False,
                            NDDEM_output=True
                            )



    positionsXYZmm = numpy.vstack([positionsXYZmm, positionXYZmmOpt])
    p_f_x = radioSphere.projectSphere.projectSphereMM(positionsXYZmm,
                                                      radiiMM[0]*numpy.ones(len(positionsXYZmm)),
                                                      sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                      pixelSizeMM=pixelSizeDetectorMM,
                                                      detectorResolution=radioMM.shape)
    residual = p_f_x - radioMM

    if GRAPH:
        plt.imshow(residual)
        plt.show()

    return positionsXYZmm, residual


def cleanDivergentScan(positionsXYZmm,radioMM,radiiMM,zoomLevel,sourceObjectDistMM,pixelSizeDetectorMM,CORxMin,CORxMax, CORxNumber,verbose=False,GRAPH=False):
    p_f_x = radioSphere.projectSphere.projectSphereMM(positionsXYZmm,
                                                      radiiMM[0]*numpy.ones(len(positionsXYZmm)),
                                                      sourceDetectorDistMM=zoomLevel*sourceObjectDistMM,
                                                      pixelSizeMM=pixelSizeDetectorMM,
                                                      detectorResolution=radioMM.shape)

    residual = p_f_x - radioMM
    print(f'Maximum |residual| {numpy.abs(residual).max()}')
    # Found an extra particle that shouldn't be there
    removed = 0
    added = 0
    # remove any overlaps
    while residual.max() > 0.9:
        positionsXYZmm, residual = removeParticle(positionsXYZmm,residual,radioMM,radiiMM,pixelSizeDetectorMM,zoomLevel,sourceObjectDistMM,verbose,GRAPH)
        removed += 1
    while residual.min() < -0.9:
        positionsXYZmm, residual = addParticle(positionsXYZmm,residual,radioMM,radiiMM,pixelSizeDetectorMM,zoomLevel,sourceObjectDistMM,CORxMin, CORxMax,CORxNumber,verbose,GRAPH)
        added +=1
    # now we clean up the cleaning up to make sure we have the right number of particles
    while added > removed:
        positionsXYZmm, residual = removeParticle(positionsXYZmm,residual,radioMM,radiiMM,pixelSizeDetectorMM,zoomLevel,sourceObjectDistMM,verbose,GRAPH)
        removed += 1
    while removed > added:
        positionsXYZmm, residual = addParticle(positionsXYZmm,residual,radioMM,radiiMM,pixelSizeDetectorMM,zoomLevel,sourceObjectDistMM,CORxMin, CORxMax,CORxNumber,verbose,GRAPH)
        added +=1

    print(f'Removed {removed} particles and added {added}.')
    return positionsXYZmm


def calculateErrors(posXYZa,posXYZb,radiiMM,verbose=True):
    distances = distance.cdist(posXYZa,posXYZb)
    distance_threshold = radiiMM[0]

    # an empty column in the distance matrix is a lost particle
    number_lost = numpy.sum((numpy.sum(distances <= distance_threshold, axis=0) == 0))
    if number_lost > 0: print(f"detectSpheres.calculateErrors(): number_lost = {number_lost}")

    # get errors
    min_errors = numpy.min(distances, axis=0) # best match for each column
    min_valid_errors = min_errors[min_errors < distance_threshold] # just where we found a particle

    err_mean    = numpy.mean(min_valid_errors)
    err_std     = numpy.std( min_valid_errors)

    return err_mean, err_std, number_lost



#!BPY

__doc__ = """
Example usage for this test.
>blender --background --factory-startup --python blender_viewer.py -- -d DATA_DIR

Notice:
'--factory-startup' is used to avoid the user default settings from
                    interfering with automated scene generation.
'--' causes blender to ignore all following arguments so python can use
them.

See blender --help for details.
"""

__author__ = "benjy marks"
__version__ = "2.3"

# This is distributed under the GPL"

import bpy, csv, sys, os
#from numpy import interp
from math import log, sqrt, atan
from mathutils import Vector


def build_world():
    print('#### Building world ####')
    scene = bpy.context.scene

    # Clear existing objects.
    scene.camera = None
    # for obj in scene.objects:
        # scene.objects.remove(obj)

    # bpy.data.worlds['World'].horizon_color=[0.0,0.0,0.0] # no backlights

def set_scene():
    scene = bpy.context.scene
    scene.render.engine = 'CYCLES'
    scene.cycles.samples = 200#0
    scene.cycles.preview_samples = 1
#     bpy.context.screen.areas[-1].spaces[0].viewport_shade = 'RENDERED'
    scene.render.use_motion_blur = True
    scene.render.motion_blur_shutter = 0.075
    scene.cycles.film_transparent = True

def make_particles(data_file_path):
    print('#### Making particles ####')

    bpy.ops.mesh.primitive_uv_sphere_add()
    ob = bpy.context.object # This is 'Sphere', others start at 'Sphere.001'

    with open(data_file_path) as f:
        reader = csv.reader(f,delimiter='\t')
        for i,row in enumerate(reader):
            if i > 0:
                particle = i
                name = 'Sphere.' + str(particle).zfill(3)
                s = float(row[3])/float(2.0)
                copy = ob.copy()
                copy.data = ob.data.copy()
                copy.scale = (s,s,s)
                # bpy.context.scene.objects.link(copy)
                bpy.data.objects[name].location = [float(row[0]),
                                                   float(row[1]),
                                                   float(row[2]),
                                                   ]

                mat = bpy.data.materials.new('M' + name)
                mat.use_nodes = True
                mat.node_tree.nodes.new(type='ShaderNodeBsdfGlass')
                mat.node_tree.nodes['Glass BSDF'].inputs['Roughness'].default_value = 0.#05
                inp = mat.node_tree.nodes['Material Output'].inputs['Surface']
                outp = mat.node_tree.nodes['Glass BSDF'].outputs['BSDF']
                mat.node_tree.links.new(inp,outp)
                bpy.data.objects[name].active_material = mat
                print('Created ' + str(i) + ' particles', end='\r')
    bpy.ops.object.shade_smooth()
    sp = bpy.data.objects["Sphere"]
    # bpy.context.scene.objects.remove(sp)

def main():
    import sys       # to get command line args
    import argparse  # to parse options for us and print a nice help message

    # get the args passed to blender after "--", all of which are ignored by
    # blender so scripts may receive their own arguments
    argv = sys.argv

    if "--" not in argv:
        argv = []  # as if no args are passed
    else:
        argv = argv[argv.index("--") + 1:]  # get all args after "--"

    # When --help or no args are given, print this help
    usage_text = \
    "Run blender in background mode with this script:"
    "  blender --background --python " + __file__ + " -- [options]"

    parser = argparse.ArgumentParser(description=usage_text)

    # Possible types are: string, int, long, choice, float and complex.
    parser.add_argument('-d', '--directory', dest='data_file_path',
                        help='This is the directory of the data files',
                        required=True, metavar='FOLDER')

    args = parser.parse_args(argv) # Ignore the args

    if not argv:
        parser.print_help()
        return
    if not args.data_file_path:
        print("Error: --text=\"some string\" argument not given, aborting.")
        parser.print_help()
        return
    data_file_path = args.data_file_path

    build_world()
    set_scene()
    make_particles(data_file_path)
    bpy.ops.object.select_all()

if __name__ == "__main__":
    main()



import numpy
import scipy.ndimage
import sys, os


def pointToDetectorPixelRange(posMM, sourceDetectorDistMM=100, pixelSizeMM=0.1, detectorResolution=[512,512]):
    """
    This function gives the detector pixel that a single point will affect.
    The main idea is that it will be used with `ROIaroundSphere` option for projectSphereMM() and in turn singleSphereToDetectorPixelRange()
    in order to only project the needed pixels.

    Parameters
    ----------
        posMM : 1x3 2D numpy array of floats
            xyz position of sphere in mm, with the origin being the middle of the detector

        sourceDetectorDistMM : float, optional
            Distance between x-ray source and middle of detector
            Set as numpy.inf for parallel projection
            Default = 100

        pixelSizeMM : float, optional
            Pixel size on detector in mm
            Default = 0.1

        detectorResolution : 2-component list of ints, optional
            Number of pixels rows, columns of detector
            Default = [512,512]

    Returns
    -------
        detectorPixel : tuple
            row, column (j,i) coordinate on detector as per figures/projectedCoords_v2.pdf
    """
    assert(len(posMM.ravel())==3)

    posMM = posMM.ravel()

    if sourceDetectorDistMM == numpy.inf:
        projectedPixelSize = 1.0
    else:
        zoomLevel = sourceDetectorDistMM/posMM[0]
        projectedPixelSize = pixelSizeMM/zoomLevel

    # This is the pixel position wrt to the middle of the detector
    YpositionProjectedPX = posMM[1] / projectedPixelSize
    ZpositionProjectedPX = posMM[2] / projectedPixelSize

    # Detector is rows, columns, so Z, Y
    detectorPX = numpy.array(detectorResolution)//2 - [ ZpositionProjectedPX, YpositionProjectedPX ]

    return numpy.round(detectorPX).astype(int)


def singleSphereToDetectorPixelRange(spherePositionMM, radiusMM, radiusMargin=0.1, sourceDetectorDistMM=100, pixelSizeMM=0.1, detectorResolution=[512,512], transformationCentreMM=None, transformationMatrix=None):
    """
    This function gives the detector pixel range that a single sphere will affect.
    The main idea is that it will be used with `ROIaroundSphere` option for projectSphereMM()
    in order to only project the needed pixels.

    Parameters
    ----------
        spherePositionMM : 1x3 2D numpy array of floats
            xyz position of sphere in mm, with the origin being the middle of the detector

        radiusMM : float
            Particle radius for projection

        radiusMargin : float
            Multiplicative margin on radius
            Default = 0.1

        ROIaroundSphere : bool, optional
            If there is only one sphere, only compute a region-of-interest radiography?
            Default = False

        sourceDetectorDistMM : float, optional
            Distance between x-ray source and middle of detector
            Set as numpy.inf for parallel projection
            Default = 100

        pixelSizeMM : float, optional
            Pixel size on detector in mm
            Default = 0.1

        detectorResolution : 2-component list of ints, optional
            Number of pixels rows, columns of detector
            Default = [512,512]

        transformationCentreMM : 3-component vector
            XYZ centre for a transformation
            Default = None

        transformationMatrix : 3x3 numpy array
            XYZ transformation matrix to apply to coordinates
            Default = None

    Returns
    -------
        JIrange : range in rows, colums of detector concerned by this grain
    """
    assert((transformationCentreMM is None) == (transformationMatrix is None)), "projectSphere.singleSphereToDetectorPixelRange(): transformationCentreMM and transformationMatrix must both be set or unset"

    # Transform coordinates if so asked
    if transformationCentreMM is not None:
        spherePositionMM = numpy.dot(transformationMatrix, numpy.array(spherePositionMM).ravel()  - numpy.array(transformationCentreMM).ravel()) + transformationCentreMM
        #spherePositionMM = numpy.array([numpy.dot(transformationMatrix, spherePositionMM[0] - transformationCentreMM) + transformationCentreMM])

    x = spherePositionMM.ravel()[0]
    y = spherePositionMM.ravel()[1]
    z = spherePositionMM.ravel()[2]

    # compute bounding square from the corners of the XYZ-aligned cube closest to the detector
    # scrap that, do all corners for safety
    corners = numpy.zeros((8,2), dtype=int)
    n = 0
    for dx in [-1, 1]:
        for dy in [-1, 1]:
            for dz in [-1, 1]:
                corners[n] = pointToDetectorPixelRange(numpy.array([x+radiusMM*dx+radiusMM*radiusMargin*dx,
                                                                    y+radiusMM*dy+radiusMM*radiusMargin*dy,
                                                                    z+radiusMM*dz+radiusMM*radiusMargin*dz]),
                                                        sourceDetectorDistMM=sourceDetectorDistMM,
                                                        pixelSizeMM=pixelSizeMM,
                                                        detectorResolution=detectorResolution)
                n += 1

    #print("singleSphereToDetectorPixelRange():", numpy.max(corners, axis=0))
    #print("singleSphereToDetectorPixelRange():", numpy.min(corners, axis=0))
    maxRC = numpy.max(corners, axis=0)
    minRC = numpy.min(corners, axis=0)

    maxRC = numpy.minimum(maxRC,detectorResolution) # clip bounding box if past limits
    minRC = numpy.maximum(minRC,[0,0]) # clip bounding box if past limits
    return numpy.array([[minRC[0], maxRC[0]], [minRC[1], maxRC[1]]])


def projectSphereMM(spheresPositionMM, radiiMM, ROIcentreMM=None, ROIradiusMM=None, sourceDetectorDistMM=100, pixelSizeMM=0.1, detectorResolution=[512,512], projector='C', transformationCentreMM=None, transformationMatrix=None, blur=None):
    """
    This is the python wrapping function for the C++ projector, it gets projection geometry,
    list of particle positions and radii, projected in the X direction.
    The output is the crossed distance for each sphere in mm.

    Please refer to the figures/projectedCoords_v2 for geometry

    In order to allow projections from diffferent angles, an XYZ centre and a transformation matrix can be provided,
    which will be applied to the particle positions.

    Parameters
    ----------
        spheresPositionMM : Nx3 2D numpy array of floats
            xyz positions of spheres in mm, with the origin being the middle of the detector

        radiiMM : 1D numpy array of floats
            Particle radii for projection

        ROIcentreMM : 3-component vector of floats, optional
            Particle position for ROI
            Default = Disactivated (None)

        ROIradiusMM : float, optional
            Particle radius for ROI
            Default = Disactivated (None)

        sourceDetectorDistMM : float, optional
            Distance between x-ray source and middle of detector.
            Set as numpy.inf for parallel projection
            Default = 100

        pixelSizeMM : float, optional
            Pixel size on detector in mm
            Default = 0.1

        detectorResolution : 2-component list of ints, optional
            Number of pixels rows, columns of detector
            Default = [512,512]

        projector : string, optional
            Algorithm for the projector (leave this alone for now)
            Default = 'C'

        transformationCentreMM : 3-component vector
            XYZ centre for a transformation
            Default = None

        transformationMatrix : 3x3 numpy array
            XYZ transformation matrix to apply to coordinates
            Default = None

        blur : float, optional
            sigma of blur to pass to scipy.ndimage.gaussian_filter to
            blur the radiograph at the end of everything

    Returns
    -------
        projectionMM : 2D numpy array of floats
            Radiography containing the total crossed distance through the spheres distance in mm for each beam path.
            To turn this into a radiography, the distances should be put into a calibrated Beer-Lambert law
    """

    assert(len(spheresPositionMM.shape) == 2),                                  "projectSphere.projectSphereMM(): spheresPositionMM is not 2D array"
    assert(len(radiiMM.shape) == 1),                                            "projectSphere.projectSphereMM(): radiiMM is not 1D array"
    assert(spheresPositionMM.shape[0] == radiiMM.shape[0]),                     "projectSphere.projectSphereMM(): number of radii and number of sphere positions not the same"
    assert((transformationCentreMM is None) == (transformationMatrix is None)), "projectSphere.projectSphereMM(): transformationCentreMM and transformationMatrix must both be set or unset"


    # Transform coordinates if so asked
    if transformationCentreMM is not None:
        tmp = spheresPositionMM - transformationCentreMM
        for n, t in enumerate(tmp):
            tmp[n] = numpy.dot(transformationMatrix, t)
        tmp += transformationCentreMM
        spheresPositionMM = tmp

        # On the fly let's also move ROIcentreMM, no this is moved into transformation options in singleSphereToDetectorPixelRange
        #if ROIcentreMM is not None:
            #ROIcentreMM = numpy.dot(transformationMatrix, numpy.array(ROIcentreMM).ravel()  - numpy.array(transformationCentreMM).ravel()) + transformationCentreMM

    # Special case: use this to indicate a parallel projection in the X-direction, so x-positions are ignored.
    if sourceDetectorDistMM == numpy.inf:
        if ROIcentreMM is None or ROIradiusMM is None:
            # Again refer to projectedCoords_v2.pdf
            # -z in space goes with j on the detector and
            # -y in space goes with i on the detector

            # use algorithm from tomopack
            iDetector = numpy.linspace( pixelSizeMM*detectorResolution[0]/2.,
                                       -pixelSizeMM*detectorResolution[0]/2.,
                                        detectorResolution[0]).astype('<f4')
            jDetector = numpy.linspace( pixelSizeMM*detectorResolution[1]/2.,
                                       -pixelSizeMM*detectorResolution[1]/2.,
                                        detectorResolution[1]).astype('<f4')

            iDetector2D, jDetector2D = numpy.meshgrid(jDetector, iDetector)

            projectionXmm = numpy.zeros(detectorResolution, dtype=('<f4'))

            for spherePositionMM, radiusMM in zip(spheresPositionMM, radiiMM):
                # This function returns the parallel projection (dims of x_detector) of particles positioned at x and y
                #print("Adding sphere at: ", spherePositionMM, 'r: ', radiusMM)
                tmp = radiusMM**2 - (spherePositionMM[1] - iDetector2D)**2    - (spherePositionMM[2] - jDetector2D)**2
                mask = tmp > 0
                projectionXmm[mask] += 2*numpy.sqrt(tmp[mask])
            return projectionXmm
        else:
            print("projectSphere.projectSphereMM(): ROI mode in parallel not yet implemented (but shoudn't be to hard)")
            return



    # Flip axes for the C++ projector after applying transformation
    spheresPositionMM = spheresPositionMM * numpy.array([1, -1, -1])

    if projector == 'C':
        #sys.path.append(os.path.join(os.path.dirname(__file__), "/"))
        import projectSphereC3

        if ROIcentreMM is None or ROIradiusMM is None:
            projectionXmm = numpy.zeros(detectorResolution, dtype=('<f4'))

            # This C++ function fill in the passed projectionXmm array
            projectSphereC3.project_func(numpy.array([sourceDetectorDistMM], dtype='<f4'),
                                         radiiMM.astype('<f4'),
                                         numpy.linspace(-pixelSizeMM*detectorResolution[0]/2.,
                                                         pixelSizeMM*detectorResolution[0]/2.,
                                                         detectorResolution[0]).astype('<f4'),
                                         numpy.linspace(-pixelSizeMM*detectorResolution[1]/2.,
                                                         pixelSizeMM*detectorResolution[1]/2.,
                                                         detectorResolution[1]).astype('<f4'),
                                         spheresPositionMM.astype('<f4'),
                                         projectionXmm)
        elif ROIcentreMM is not None and ROIradiusMM is not None:
            # Make sure there's only one sphere:
            assert(len(spheresPositionMM.ravel())==3), "projectSphere.projectSphereMM(): in ROI mode I want only one sphere"

            # Get limits
            limits = singleSphereToDetectorPixelRange(ROIcentreMM.ravel(),
                                                      ROIradiusMM,
                                                      radiusMargin=0.1,
                                                      sourceDetectorDistMM=sourceDetectorDistMM,
                                                      pixelSizeMM=pixelSizeMM,
                                                      detectorResolution=detectorResolution,
                                                      transformationCentreMM=transformationCentreMM,
                                                      transformationMatrix=transformationMatrix)

            # Define (smaller) projection array to fill in
            projectionXmm = numpy.zeros((limits[0,1]-limits[0,0], limits[1,1]-limits[1,0]), dtype=('<f4'))

            # This C++ function fill in the passed projectionXmm array -- not limits after linspace
            projectSphereC3.project_func(numpy.array([sourceDetectorDistMM], dtype='<f4'),
                                         radiiMM.astype('<f4'),
                                         numpy.linspace(-pixelSizeMM*detectorResolution[0]/2.,
                                                         pixelSizeMM*detectorResolution[0]/2.,
                                                         detectorResolution[0]).astype('<f4')[limits[0,0]:limits[0,1]],
                                         numpy.linspace(-pixelSizeMM*detectorResolution[1]/2.,
                                                         pixelSizeMM*detectorResolution[1]/2.,
                                                         detectorResolution[1]).astype('<f4')[limits[1,0]:limits[1,1]],
                                         spheresPositionMM.astype('<f4'),
                                         projectionXmm)
        else:
            print("projectSphere.projectSphereMM(): If you set ROIcentreMM you must also set ROIradiusMM")

    else:
        print("projectSphere.projectSphereMM(): This projection mode not implemented")
        return

    if blur is not None:
        projectionXmm = scipy.ndimage.gaussian_filter(projectionXmm, sigma=blur)

    return projectionXmm


def computeLinearBackground(radioMM, mask=None):
    """
    This function computes a plane-fit for background greylevels to help with the correction of the background


    Parameters
    ----------
        radioMM : 2D numpy array of floats
            The image

        mask : 2D numpy array of bools, optional
            Masked zone to fit?

    Returns
    -------
        background : 2D numpy array of floats
            Same size as radioMM
    """
    x,y = numpy.meshgrid(numpy.arange(radioMM.shape[0]), numpy.arange(radioMM.shape[1]), indexing='ij')
    def plane(params,z,validPoints):
        output = params[0]*x[validPoints] + params[1]*y[validPoints] + z[validPoints] - params[2]
        return numpy.ravel(output)

    if mask is None:
        mask = numpy.abs(radioMM) < 0.05*radiiMM[0] # just where it is reasonable valued
    #isBackgroundMask *= numpy.abs(residualMM) > 0.1*radiiMM[0] # just where it is reasonable valued
    #plt.imshow(isBackgroundMask); plt.show()
    LSQret = scipy.optimize.least_squares(plane,
                                            [0,0,0],
                                            args=[radioMM,mask])

    backgroundPlane = - LSQret['x'][0]*x - LSQret['x'][1]*y + LSQret['x'][2]

    return backgroundPlane


def gl2mm(radio, calib=None):
    """
    This function takes a greylevel radiograph (I/I0) and returns
    a radiograph in mm (L), representing the path length encountered.

    Parameters
    ----------
        radio : a 2D numpy array of floats

        calib : dictionary (optional)
            This contains a calibration of greylevels to mm
            If not passed \mu is assumed to be 1.
    """
    if calib is not None:
        print("projectSphere.gl2mm() I need to be implemented")
    return -numpy.log(radio)


def mm2gl(radioMM, calib=None):
    """
    This function takes a radiograph in mm (L) and returns
    a radiograph in greylevels (I/I0)

    Parameters
    ----------
        radioMM : a 2D numpy array of floats

        calib : dictionary (optional)
            This contains a calibration of greylevels to mm
            If not passed \mu is assumed to be 1.
    """
    if calib is not None:
        print("projectSphere.mm2gl() I need to be implemented")
    return numpy.exp(-radioMM)



import numpy
import tifffile
import sys, os
# sys.path.append(os.path.join(os.path.dirname(__file__), "projectSphereC.cython"))
import radioSphere.projectSphere
import csv

def getParticlesFromMercury(mercuryFileName, timeStepDEM):
    with open(mercuryFileName) as f:
        frame = 0
        reader = csv.reader(f,delimiter=' ')
        for i,line in enumerate(reader):
            if i == 0:
                num_particles = int(line[0])
                loc = numpy.zeros([num_particles, 3], dtype="<f4")
                radii = numpy.zeros([num_particles],  dtype="<f4")
            elif i%(num_particles+1) == 0:
                frame += 1
                # print('Made ' + str(frame).zfill(3) + ' frames', end='\r')
                if frame > timeStepDEM:
                    break
            elif frame == timeStepDEM:
                pt = i%(num_particles+1) - 1
                loc[pt,0] = numpy.double(line[0]) # x
                loc[pt,1] = numpy.double(line[1]) # y
                loc[pt,2] = numpy.double(line[2]) # z
                radii[pt] = numpy.double(line[6]) # radius
    return loc, radii

def generateSampledRadiographFromMercuryData(mercuryFileName, timeStepDEM, halfConeBeamAngleDegrees, pixelSizeMM, detectorResolution, samplingThicknessXMM=None, samplingTransverseDistanceMM=None, samplingFraction=None, zoomLevel=2.0, saveImage=False, GRAPH=False):
    # Load data
    spheres, radii = getParticlesFromMercury(mercuryFileName,timeStepDEM)
    spheresMM = spheres*1000
    radiiMM = radii*1000

    if samplingThicknessXMM is not None:
        # just sample in the middle samplingThicknessXMM in the X direction
        sampled = ( spheresMM[:,0] > -samplingThicknessXMM/2. ) * ( spheresMM[:,0] < samplingThicknessXMM/2. )
    if samplingFraction is not None:
         # sampling_parameter is volume fraction of solids (i.e. fraction of spheres in packing)
        sampled = numpy.random.choice(len(spheresMM), int(samplingFraction*len(spheresMM)))
    spheresMM = spheresMM[sampled,:]
    radiiMM = radiiMM[sampled]

    if samplingTransverseDistanceMM is not None:
        sampled = numpy.sqrt(spheresMM[:,1]**2 + spheresMM[:,2]**2) < samplingTransverseDistanceMM
        spheresMM = spheresMM[sampled,:]
        radiiMM = radiiMM[sampled]

    sourceObjectDistMM = radiiMM[0]/numpy.tan(numpy.radians(halfConeBeamAngleDegrees))
    # spheresMM[:,0] -= spheresMM[:,0].mean()
    spheresMM[:,0] += sourceObjectDistMM # move to sample location

    pixelsPerDiameterZoomed = radiiMM*2 / (pixelSizeMM / zoomLevel)
    sourceDetectorDistMM = zoomLevel * sourceObjectDistMM

    radioMM = radioSphere.projectSphere.projectSphereMM(spheresMM,
                                                        radiiMM,
                                                        detectorResolution=detectorResolution,
                                                        pixelSizeMM=pixelSizeMM,
                                                        sourceDetectorDistMM=sourceDetectorDistMM)

    if GRAPH:
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

        # Functions from @Mateen Ulhaq and @karlo
        def set_axes_equal(ax: plt.Axes):
            limits = numpy.array([
                ax.get_xlim3d(),
                ax.get_ylim3d(),
                ax.get_zlim3d(),
            ])
            origin = numpy.mean(limits, axis=1)
            radius = 0.5 * numpy.max(numpy.abs(limits[:, 1] - limits[:, 0]))
            _set_axes_radius(ax, origin, radius)

        def _set_axes_radius(ax, origin, radius):
            x, y, z = origin
            ax.set_xlim3d([x - radius, x + radius])
            ax.set_ylim3d([y - radius, y + radius])
            ax.set_zlim3d([z - radius, z + radius])

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(spheresMM[:,0], spheresMM[:,1], spheresMM[:,2], marker='o', s=200)
        ax.set_box_aspect([1,1,1]) # IMPORTANT - this is the new, key line
        # ax.set_proj_type('ortho') # OPTIONAL - default is perspective (shown in image above)
        set_axes_equal(ax) # IMPORTANT - this is also required        plt.show()
        plt.show()

    if saveImage:
        # print('saving')
        foldername = 'data/DEM_projected/half-cone-beam-angle-' + str(halfConeBeamAngleDegrees) + \
                     '/resolution-' + str(pixelsPerDiameterZoomed) + \
                     '/thickness-' + str(numberOfParticleLayers) + '/'
        if not os.path.exists(foldername): os.makedirs(foldername)
        tifffile.imsave(foldername + str(tstep).zfill(4)+'.tiff',radioXmm)

    return radioMM, spheresMM, radiiMM, sourceObjectDistMM


if __name__ == '__main__':
    fname = './data/lees_edwards_nu_0.6_R_1_M_1.data'
    # Constant system parameters
    particleDiameterMM   = 1 # Set in DEM
    pixelSizeMM          = 0.1
    detectorResolution   = [512,512]
    maximumAspectRatio   = 1.5

    for tstep in range(1):
        for halfConeBeamAngleDegrees in [1,2,3,5,10]:
            for numberOfParticleLayers in [0.01,0.5,1,2,5,7,10]:
                generate_radiograph_from_DEM(fname,tstep,halfConeBeamAngleDegrees,numberOfParticleLayers,particleDiameterMM,pixelSizeMM,detectorResolution, samplingMode='layers')



import math
import numpy
import matplotlib.pyplot as plt
import tifffile
from scipy.optimize import curve_fit


def cubicFit( x, l, n, m, c ): return l*x**3 + n*x**2 + m*x + c
def quadraticFit( x, n, m, c ): return n*x**2 + m*x + c
def linearFit( x, m, c ): return m*x + c
def linearFitNoIntercept( x, m ): return m*x

def getPathLengthThroughSphere(p, r, sdd, sod):
    '''
    Get the path length of a ray passing through a sphere. Schematic diagram is provided in the supplementary methods section of the first (only?) radiosphere paper.

    Parameters
    ----------
        p : float
            The distance in mm along the detector panel from the centre.

        r : float
            The particle radius in mm.

        sdd : float
            The source-detector distance in mm.

        sod : float
            The source-object distance in mm.

    Returns
    -------
        L : float
            The path length through the sphere in mm. If the path lies outside the sphere, returns 0.
    '''
    try:
        p = float(p)
        #print p,
        alpha = math.atan( p/sdd )
        #print alpha
        beta = math.asin( sod*math.sin(alpha) / r  )
        L = 2.0*r*math.sin(math.pi/2.0-beta)
        #print numpy.rad2deg(alpha), numpy.rad2deg(beta), L
        return L
    except:
        return 0.0

def generateFitParameters(calibSphereLog,pixelSizeMM,sourceDetectorDistMM,sourceObjectDistanceMM,radiusMM,centreYXpx,fitFunction=linearFit,outputPath=False,verbose=False):
    '''
    Fit an attenuation law to a logged normalised radiograph of a single particle.

    Parameters
    ----------
        calibSphereLog : 2D numpy array of floats
            A radiograph of a single particle divided by the background intensity and then logged.

        pixelSizeMM : float
            The size of a single pixel in mm on the detector panel.

        sourceDetectorDistMM: float
            The source-detector distance in mm.

        sourceObjectDistanceMM: float
            The source-object distance in mm.

        radiusMM : float
            The particle radius in mm.

        centreYXpx : 1D numpy array of floats
             The y and x location of the centre of the particle in pixels.

        fitFunction : function handle (optional)
            The fitting function to use. By default fit a linear fitting law. Options are `cubicFit`, `quadraticFit`, `linearFit` and `linearFitNoIntercept`

        outputPath : string (optional)
            A path to save the fit as an `npy` file.

        verbose : bool (optional)
            Show the fit on a graph. Default is False.

    Returns
    -------
        L: float
            The path length through the sphere in mm. If the path lies outside the sphere, returns 0.
    '''
    alphaMax = math.asin( radiusMM/sourceObjectDistanceMM )
    pmax = math.tan( alphaMax ) * sourceDetectorDistMM

    # Pixel positions going -x from the middle of the sphere in MM
    pixelPoints   = numpy.array(range(int(round(centreYXpx[0])))[::-1])
    pixelPointsMMdetector = pixelPoints*pixelSizeMM


    points = []
    #for pn, pMMdet in enumerate(pixelPointsMMdetector[0:130]):
    for pn, pMMdet in enumerate(pixelPointsMMdetector):
        #if pMMdet < pmax:
        L = getPathLengthThroughSphere( pMMdet, radiusMM, sourceDetectorDistMM, sourceObjectDistanceMM )
        if L > 0:
            points.append( [ L, calibSphereLog[pixelPoints[-pn], int(centreYXpx[1])], pixelPoints[-pn] ] )
    points = numpy.array( points )

    poptN,pcov  = curve_fit(fitFunction, points[:,1], points[:,0])

    if outputPath: numpy.save(outputPath, poptN)

    if verbose:
        D = 150
        plt.subplot(121)
        plt.imshow(calibSphereLog)
        plt.colorbar()
        plt.plot(int(centreYXpx[1])*numpy.ones_like(points[:,2]),points[:,2],'w--')
        plt.plot(int(centreYXpx[1]),points[0,2],'wx')
        plt.plot(int(centreYXpx[1]),points[-1,2],'wx')
        plt.xlim(centreYXpx[1]-D,centreYXpx[1]+D)
        plt.ylim(centreYXpx[0]-D,centreYXpx[0]+D)
        plt.subplot(122)
        plt.plot( points[:,1], points[:,0], 'k.', label='Measured value' ) # Measured calib sphere with 130kV\nand 1.00mm Cu filter
        plt.plot( points[:,1], fitFunction( points[:,1], *poptN ), 'k-', alpha=0.5, label='Fit' )
        plt.ylabel("Path length inside\nsphere (mm)")
        # plt.ylim([0, max(points[:,0])])
        plt.xlabel(r"Log Attenuation $\ln(I/I_0)$")
        plt.legend(loc=0)

        # plt.subplots_adjust(bottom=0.21,top=0.99,right=0.99,left=0.16)
        plt.show()
        # plt.savefig('./figures/experimental-attenuationCalibration.pdf')


    return poptN

if __name__ == '__main__':
    plt.style.use('./tools/radioSphere.mplstyle')

    calibSpherePath = './data/2021-02-09-EddyBillesNanoBis/7mm-AVG64.tif'
    backgroundPath = './data/2021-02-09-EddyBillesNanoBis/I0-AVG64.tif'
    outputPath = "./cache/fit-log-linear.npy"

    # Load images
    calibSphere = tifffile.imread(calibSpherePath).astype(float) / tifffile.imread(backgroundPath).astype(float)
    calibSphereLog = numpy.log(calibSphere)

    # Projection geometry stuff
    binning=4
    pixelSizeMM = 0.127*float(binning)
    sourceDetectorDistMM = 242.597 # from XAct
    radiusMM = 7/2
    sourceObjectDistanceMM  = sourceDetectorDistMM * ( radiusMM/71 / pixelSizeMM)     # 71 pixels across diameter, so 51um/px, pixels 0.508 mm
    # uncertain parameter, this wiggles it
    sourceObjectDistanceMM += 0.5
    # print("SOD = ", sourceObjectDistanceMM)
    centreYXpx = numpy.array([ 229, 183 ])

    poptN = generateFitParameters(calibSphereLog,pixelSizeMM,sourceDetectorDistMM,sourceObjectDistanceMM,radiusMM,centreYXpx,outputPath=outputPath,verbose=True)
